#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "schrodinger2d.h"
#include <pthread.h>
#include <GL/glew.h>
#include <GL/freeglut_std.h>
#include <GL/freeglut_ext.h>

#include "shaders.h"
#include "transfo.h"

 
#ifdef OPENCL
#include <CL/cl.h>
#include <CL/cl_gl.h>
//#include <GL/glx.h>
#define PROBA_LOCAL_WORK_SIZE 16
#endif

typedef unsigned char uint8_t;
typedef unsigned int uint32_t;


#define IND(i,j) (i)+(j)*solver.full_width



typedef struct {
    char type;
    int px_min,py_min;
    int levels;
    int scale;
    float Lx,Ly;
    short width,height;// taille de la grille (nombre de points)
    short full_width,full_height; // taille de la grille complète (avec les rangées fictives)
    int size, full_size; // taille total des tableaux
    float dx,dy,dxdy; // pas de la grille
    float *U_re,*U_im,*U1_im,*U1_re,*P;
    float *A,*B_re,*B_im,*C,*D,*E_re,*E_im,*F,*G;
    float *beta_re,*beta_im, *beta_r2, *gamma_re,*gamma_im;
    float *beta_p_re,*beta_p_im, *beta_p_r2, *gamma_p_re,*gamma_p_im;
    float *X_re,*X_im;
    float *V;
    float *contours;
    float t,dt,ti,tf;
    float alpha;
    int normalize,colormap;
    float *image;
    float *map_r,*map_g,*map_b;
    float cont_r, cont_g, cont_b;
    short cont;
    int surface;
    float *sommets_xy, *sommets_z, *normales;
    float x0,y0;
    uint32_t *indices;
    float surface_height;
    uint8_t *render_img;
    int win_width,win_height;
    
    int glut_init;
    GLuint gl_program_image,gl_program_surface,gl_program_pixmap_modelview;
    GLuint tex_r, tex_g, tex_b, tex_rgb;
    
    int nb_sommets;
    GLuint buffer_indices, buffer_sommets_xy, buffer_sommets_z, buffer_normales, buffer_colors;
    float LightPosition[4];
    float Kdiffuse[3];
    float LightIntensity[3];
    float Kambient[3];
    Transformation *transfo_P, *transfo_MV, *transfo_PMV;
    float Nmatrix[9];

#ifdef OPENCL
    float *Pmax;
    int Pmax_size_x, Pmax_size_y, Pmax_size;
    cl_mem gpu_U_re, gpu_U_im, gpu_U1_re, gpu_U1_im, gpu_P, gpu_Pmax;
    cl_mem gpu_A, gpu_D, gpu_E_re, gpu_E_im, gpu_F, gpu_G;
    cl_mem gpu_beta_re, gpu_beta_im, gpu_beta_r2, gpu_gamma_re, gpu_gamma_im, gpu_beta_p_re, gpu_beta_p_im, gpu_beta_p_r2, gpu_gamma_p_re, gpu_gamma_p_im;
    cl_mem gpu_map_r, gpu_map_g, gpu_map_b;
    cl_mem gpu_contours;
    cl_mem gpu_image;
    cl_mem gpu_sommets_z, gpu_normales;
#endif
    
} solver_def;

solver_def solver;

pthread_barrier_t threadBarrier;
typedef struct {
    int niter;
    int id;
    int iStart;
    int iEnd;
    int jStart;
    int jEnd;
    float *U_re,*U_im;
    float *U1_re,*U1_im;
} threadArg;





int schrodinger2d_open(char type,int px_min, int py_min, int levels, float Lx, int normalize, int colormap, float gamma) {
    int k,i,j,k2;
    float T,r,g,b,L,x0,y0;
    solver.type = type;
    solver.px_min = px_min;
    solver.py_min = py_min;
    solver.levels = levels;
    solver.width = (int)pow(2,px_min+levels);
    solver.height = (int)pow(2,py_min+levels);
    solver.scale = (int)pow(2,levels);
    solver.full_width = solver.width+2;
    solver.full_height = solver.height+2;
    solver.size = solver.width * solver.height;
    solver.full_size = solver.full_width * solver.full_height;
    solver.dx = Lx/(solver.width-1);
    solver.Lx = Lx;
    solver.dy = solver.dx;
    solver.dxdy = solver.dx*solver.dy;
    solver.U_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.U1_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.U_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.U1_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.P = (float*) malloc(sizeof(float)*solver.full_size);
    solver.A  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.B_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.B_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.C  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.D  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.E_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.E_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.F  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.G  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.beta_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.beta_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.beta_r2 = (float*) malloc(sizeof(float)*solver.full_size);
    solver.gamma_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.gamma_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.beta_p_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.beta_p_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.beta_p_r2 = (float*) malloc(sizeof(float)*solver.full_size);
    solver.gamma_p_re  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.gamma_p_im  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.X_re = (float*) malloc(sizeof(float)*solver.full_size);
    solver.X_im = (float*) malloc(sizeof(float)*solver.full_size);
    solver.V  = (float*) malloc(sizeof(float)*solver.full_size);
    solver.contours  = (float*) malloc(sizeof(float)*solver.size);
    
    solver.t = 0;
    for (k=0; k<solver.full_size; k++) {
        solver.U_re[k] = 0;
        solver.U_im[k] = 0;
        solver.V[k] = 0;   
    }
    for (k=0; k<solver.size; k++) solver.contours[k] = 0;
    solver.normalize = normalize;
    solver.map_r = (float*) malloc(sizeof(float)*256);
    solver.map_g = (float*) malloc(sizeof(float)*256);
    solver.map_b = (float*) malloc(sizeof(float)*256);
    solver.image = (float*) malloc(4*sizeof(float)*solver.size);
    if (colormap==COLORMAP_TSV) {
        for (k=0; k<256; k++) {
               T= (1.0-k*1.0/255)*4;
               if (T>5) {r=1; g=0; b=6-T;}
               else if (T>4) {b=1; g=0; r=T-4;}
               else if (T>3) {b=1; r=0; g=4-T;}
               else if (T>2) {g=1; r=0; b=T-2;}
               else if (T>1) {g=1; b=0; r=2-T;}
               else {r=1; b=0; g=T;}
               L = pow(k*1.0/256,gamma);
               solver.map_r[k] = r*L;
               solver.map_g[k] = g*L;
               solver.map_b[k] = b*L;
        }
    }
    else if (colormap==COLORMAP_GRAY) {
         for (k=0; k<256; k++) {
            L = pow(k*1.0/256,gamma);
            solver.map_r[k] = L;
            solver.map_g[k] = L;
            solver.map_b[k] = L;
         }
    }
    
    solver.colormap = colormap;
    solver.cont_r = 1.0;
    solver.cont_g = 1.0;
    solver.cont_b = 1.0;
    solver.cont = 1;
    
    solver.surface = 1;
    solver.sommets_xy = (float*) malloc(2*sizeof(float)*solver.size);
    solver.sommets_z = (float*)malloc(sizeof(float)*solver.size);
    x0 = -solver.Lx*0.5;
    y0 = -solver.Lx*0.5*solver.height*1.0/solver.width;
    solver.x0 = x0;
    solver.y0 = y0;
    k = 0;
    for (j=0; j<solver.height; j++) {
        for (i=0; i<solver.width; i++) {
            solver.sommets_xy[k++] = x0+i*solver.dx;
            solver.sommets_xy[k++] = y0+j*solver.dy;
        }
    }
    solver.normales = (float*) malloc(3*sizeof(float)*solver.size);
    solver.nb_sommets = (solver.width-1)*(solver.height-1)*6;
    solver.indices = (uint32_t*) malloc(sizeof(uint32_t)*solver.nb_sommets);
    k = 0; k2=0;
    for (j=0; j<solver.height-1; j++) {
        for (i=0; i<solver.width-1; i++) {
            solver.indices[k2++] = k;
            solver.indices[k2++] = k+1;
            solver.indices[k2++] = k+solver.width;
            solver.indices[k2++] = k+solver.width;
            solver.indices[k2++] = k+1;
            solver.indices[k2++] = k+solver.width+1;
            k++;
        }
        k++;
    }
    
    
    transfo_new(&solver.transfo_P);
    transfo_new(&solver.transfo_PMV);
    transfo_new(&solver.transfo_MV);
    
    solver.glut_init = 0;
    
#ifdef OPENCL
    solver.Pmax_size_x = solver.width / PROBA_LOCAL_WORK_SIZE;
    solver.Pmax_size_y = solver.height / PROBA_LOCAL_WORK_SIZE;
    solver.Pmax_size = solver.Pmax_size_x*solver.Pmax_size_y;
    solver.Pmax = (float*) malloc(sizeof(float)*solver.Pmax_size);
#endif
    return 0;
}

int schrodinger2d_close(void) {
    free(solver.U_re);
    free(solver.U1_re);
    free(solver.U_im);
    free(solver.U1_im);
    free(solver.P);
    free(solver.A);
    free(solver.B_re);
    free(solver.B_im);
    free(solver.C);
    free(solver.D);
    free(solver.E_re);
    free(solver.E_im);
    free(solver.F);
    free(solver.G);
    free(solver.gamma_re);
    free(solver.gamma_im);
    free(solver.beta_re);
    free(solver.beta_im);
    free(solver.beta_r2);
    free(solver.gamma_p_re);
    free(solver.gamma_p_im);
    free(solver.beta_p_re);
    free(solver.beta_p_im);
    free(solver.beta_p_r2);
    free(solver.X_re);
    free(solver.X_im);
    free(solver.V);
    free(solver.contours);
    free(solver.map_r);
    free(solver.map_g);
    free(solver.map_b);
    free(solver.image);
    free(solver.sommets_xy);
    free(solver.sommets_z);
    free(solver.normales);
    free(solver.indices);
    transfo_delete(solver.transfo_P);
    transfo_delete(solver.transfo_PMV);
    transfo_delete(solver.transfo_MV);
#ifdef OPENCL
    free(solver.Pmax);
#endif
    return 0;   
}

int schrodinger2d_time_step(float dt) {
    solver.dt = dt;
    solver.alpha = 2*solver.dx*solver.dx/solver.dt;
    return 0;
}

int schrodinger2d_potential_rectangle(int ci0, int cj0, int width, int height, float v) {
    int ci,cj;
    int w,h;
    int i,j;
    int k,k0;
    ci = ci0*solver.scale;
    cj = cj0*solver.scale;
    w = width*solver.scale;
    h = height*solver.scale;
    for (i=-w; i<=w; i++)
        for (j=-h; j<=h; j++)
            {
                k = ci+i+1+(cj+j+1)*solver.full_width;
                if ((k>0)&&(k<solver.full_size)) {
                    solver.V[k] = v;
                    k0 = ci+i+(cj+j)*solver.width;
                    solver.contours[k0] = 1.0;
                }
            }
    return 0;   
}

int schrodinger2d_potential_disk(int ci0, int cj0, int radius, float v) {
    int ci,cj,r;
    int i,j,imax,k,k0;
    ci = ci0*solver.scale;
    cj = cj0*solver.scale;
    r = solver.scale*radius;
    
    for (j=-r; j<=r; j++) {
           imax = sqrt(r*r-j*j);
           for (i=-imax; i<=imax; i++) {
                k = ci+i+1+(cj+j+1)*solver.full_width;
                if ((k>0)&&(k<solver.full_size)) {
                    solver.V[k] = v;
                    k0 = ci+i+(cj+j)*solver.width;
                    solver.contours[k0] = 1.0;
                }
           }
    }
    return 0;
}

int schrodinger2d_potential(float *data) {
    int k1,k2,j,size;
    k1 = solver.full_width+1;
    k2 = 0;
    size = solver.width*sizeof(float);
    for (j=1; j<=solver.height; j++) {
        memcpy(solver.V+k1,data+k2,size);
        k1 += solver.full_width;
        k2 += solver.width;
    }
    return 0;
}

int _schrodinger(void) {
    int k;
    float a;
    float v;
    a = solver.alpha;
    for (k=0; k<solver.full_size; k++) {
        v = 0.5*solver.dx*solver.dx*solver.V[k];
        solver.A[k] = 1.0;
        solver.B_re[k] = -2.0-v;
        solver.B_im[k] = a;
        solver.C[k] = 1.0;
        solver.D[k] = -1.0;
        solver.E_re[k] = 2.0+v;
        solver.E_im[k] = a;
        solver.F[k] = -1.0;
        solver.G[k] = 0.0;
        
    }
    return 0;
}

int schrodinger2d_schrodinger() {
    switch (solver.type) {
    case TYPE_CARTESIAN:_schrodinger();   
    }
    return 0;
}

int _dirichlet_borders(void) {
       int i,j;
       int k;
       for (i=1; i<=solver.width; i++) {
            k = IND(i,1);
            solver.A[k] = 0;
            solver.B_re[k] = 1;
            solver.B_im[k] = 0;
            solver.C[k] = 0;
            solver.D[k] = 0;
            solver.E_re[k] = 0;
            solver.E_im[k] = 0;
            solver.F[k] = 0;
            solver.G[k] = 0;
            solver.U_re[k] = 0;
            solver.U_im[k] = 0;
            
            k = IND(i,solver.height);
            solver.A[k] = 0;
            solver.B_re[k] = 1;
            solver.B_im[k] = 0;
            solver.C[k] = 0;
            solver.D[k] = 0;
            solver.E_re[k] = 0;
            solver.E_im[k] = 0;
            solver.F[k] = 0;
            solver.G[k] = 0;
            solver.U_re[k] = 0;
            solver.U_im[k] = 0;
            
       }
       for (j=1; j<=solver.height; j++) {
            k = IND(1,j);
            solver.A[k] = 0;
            solver.B_re[k] = 1;
            solver.B_im[k] = 0;
            solver.C[k] = 0;
            solver.D[k] = 0;
            solver.E_re[k] = 0;
            solver.E_im[k] = 0;
            solver.F[k] = 0;
            solver.G[k] = 0;
            solver.U_re[k] = 0;
            solver.U_im[k] = 0;
            
            k = IND(solver.width,j);
            solver.A[k] = 0;
            solver.B_re[k] = 1;
            solver.B_im[k] = 0;
            solver.C[k] = 0;
            solver.D[k] = 0;
            solver.E_re[k] = 0;
            solver.E_im[k] = 0;
            solver.F[k] = 0;
            solver.G[k] = 0;
            solver.U_re[k] = 0;
            solver.U_im[k] = 0;
            
       }
       return 0;
}

int schrodinger2d_dirichlet_borders() {
    switch (solver.type) {
    case TYPE_CARTESIAN: _dirichlet_borders();   
    }
    return 0;
}

int _dirichlet_point(int i, int j) {
    int k,k0;
    if ((i<0)||(i>solver.width)||(j<0)||(j>solver.height)) return 0;
    k = i+1+(j+1)*solver.full_width;
    solver.A[k] = 0;
    solver.B_re[k] = 1;
    solver.B_im[k] = 0;
    solver.C[k] = 0;
    solver.D[k] = 0;
    solver.E_re[k] = 0;
    solver.E_im[k] = 0;
    solver.F[k] = 0;
    solver.G[k] = 0;
    solver.U_re[k] = 0;
    solver.U_im[k] = 0;
    k0 = i+j*solver.width;
    solver.contours[k0] = 1.0;
    return 0;   
}

int schrodinger2d_dirichlet_line(int i0, int j0, int length, int direction) {
    int istart,jstart;
    int L;
    int i,j;
    istart = i0*solver.scale;
    jstart = j0*solver.scale;
    L = length*solver.scale;
    switch(direction) {
    case LINE_UP: for (j=jstart; j<jstart+L+1; j++) _dirichlet_point(istart,j); break;
    case LINE_DOWN :for (j=jstart; j>jstart-L-1; j--) _dirichlet_point(istart,j); break;
    case LINE_LEFT : for (i=istart; i<istart+L+1; i++) _dirichlet_point(i,jstart); break;
    case LINE_RIGHT : for (i=istart; i>istart-L-1; i--) _dirichlet_point(i,jstart); break;
    }
    return 0;
}

int _init_mat(void) {
    int i,j,k,kn;
    float r2;
    for (j=1; j<=solver.height; j++) {
        k = IND(1,j);
        solver.beta_re[k] = solver.B_re[k];
        solver.beta_im[k] = solver.B_im[k];
        r2 = solver.beta_re[k]*solver.beta_re[k]+solver.beta_im[k]*solver.beta_im[k];
        solver.gamma_re[k] = solver.C[k]*solver.beta_re[k]/r2;
        solver.gamma_im[k] = -solver.C[k]*solver.beta_im[k]/r2;
        solver.beta_r2[k] = r2;
        kn = k;
        for (i=2; i<=solver.width; i++) {
               k++;
               solver.beta_re[k] = solver.B_re[k]-solver.A[k]*solver.gamma_re[kn];
               solver.beta_im[k] = solver.B_im[k]-solver.A[k]*solver.gamma_im[kn];
               r2 = solver.beta_re[k]*solver.beta_re[k]+solver.beta_im[k]*solver.beta_im[k];
               solver.gamma_re[k] = solver.C[k]*solver.beta_re[k]/r2;
               solver.gamma_im[k] = -solver.C[k]*solver.beta_im[k]/r2;
               solver.beta_r2[k] = r2;
               kn = k;
        }
    }
    for (i=1; i<=solver.width; i++) {
        k = IND(i,1);
        solver.beta_p_re[k] = solver.B_re[k];
        solver.beta_p_im[k] = solver.B_im[k];
        r2 = solver.beta_p_re[k]*solver.beta_p_re[k]+solver.beta_p_im[k]*solver.beta_p_im[k];
        solver.gamma_p_re[k] = solver.C[k]*solver.beta_p_re[k]/r2;
        solver.gamma_p_im[k] = -solver.C[k]*solver.beta_p_im[k]/r2;
        solver.beta_p_r2[k] = r2;
        kn = k;
        for (j=2; j<=solver.height; j++) {
            k += solver.full_width;
            solver.beta_p_re[k] = solver.B_re[k]-solver.A[k]*solver.gamma_p_re[kn];
            solver.beta_p_im[k] = solver.B_im[k]-solver.A[k]*solver.gamma_p_im[kn];
            r2 = solver.beta_p_re[k]*solver.beta_p_re[k]+solver.beta_p_im[k]*solver.beta_p_im[k];
            solver.gamma_p_re[k] = solver.C[k]*solver.beta_p_re[k]/r2;
            solver.gamma_p_im[k] = -solver.C[k]*solver.beta_p_im[k]/r2;
            solver.beta_p_r2[k] = r2;
            kn = k;
        }
    }
    return 0;
}

int schrodinger2d_init() {
    switch (solver.type) {
    case TYPE_CARTESIAN: _init_mat();   
    }
    return 0;
}



int _row_iteration(int j, float *U_re, float *U_im, float *U1_re, float *U1_im) {
    int k,kn;
    int i;
    float r2;
    float a,b;
    i = 1;
    k = IND(i,j);
    a = solver.D[k]*U_re[k-solver.full_width]+solver.E_re[k]*solver.U_re[k]-solver.E_im[k]*solver.U_im[k]+solver.F[k]*U_re[k+solver.full_width]+solver.G[k];
    b = solver.D[k]*U_im[k-solver.full_width]+solver.E_re[k]*solver.U_im[k]+solver.E_im[k]*solver.U_re[k]+solver.F[k]*U_im[k+solver.full_width]+solver.G[k];
    //solver.X[k] = (solver.D[k]*U[k-solver.full_width]+solver.E[k]*U[k]+solver.F[k]*U[k+solver.full_width]+solver.G[k])/solver.beta[k];
    r2 = solver.beta_r2[k];
    solver.X_re[k] = (a*solver.beta_re[k]+b*solver.beta_im[k])/r2;
    solver.X_im[k] = (b*solver.beta_re[k]-a*solver.beta_im[k])/r2;
    kn = k;
    for (i=2; i<=solver.width; i++) {
        k++;
        //solver.X[k] = (solver.D[k]*U[k-solver.full_width]+solver.E[k]*U[k]+solver.F[k]*U[k+solver.full_width]+solver.G[k]-solver.A[k]*solver.X[k-1])/solver.beta[k];
        a = solver.D[k]*U_re[k-solver.full_width]+solver.E_re[k]*solver.U_re[k]-solver.E_im[k]*solver.U_im[k]+solver.F[k]*U_re[k+solver.full_width]+solver.G[k]-solver.A[k]*solver.X_re[kn];
        b = solver.D[k]*U_im[k-solver.full_width]+solver.E_re[k]*solver.U_im[k]+solver.E_im[k]*solver.U_re[k]+solver.F[k]*U_im[k+solver.full_width]+solver.G[k]-solver.A[k]*solver.X_im[kn];
        r2 = solver.beta_r2[k];
        solver.X_re[k] = (a*solver.beta_re[k]+b*solver.beta_im[k])/r2;
        solver.X_im[k] = (b*solver.beta_re[k]-a*solver.beta_im[k])/r2;
        kn = k;
    }
    i = solver.width;
    k = IND(solver.width,j);
    kn = k;
    U1_re[k] = solver.X_re[k];
    U1_im[k] = solver.X_im[k];
    while (i>1) {
        k--;
        i--;
        //U1[k] = solver.X[k]-solver.gamma[k]*U1[kn];
        U1_re[k] = solver.X_re[k]-(solver.gamma_re[k]*U1_re[kn]-solver.gamma_im[k]*U1_im[kn]);
        U1_im[k] = solver.X_im[k]-(solver.gamma_re[k]*U1_im[kn]+solver.gamma_im[k]*U1_re[kn]);
        kn = k;
    }
    return 0;
}

int _col_iteration(int i, float *U_re, float *U_im, float *U1_re, float *U1_im) {
    int k,kn;
    int j;
    float r2;
    float a,b;
    j = 1;
    k = IND(i,j);
    //solver.X[k] = (solver.D_p[k]*U[k-1]+solver.E_p[k]*U[k]+solver.F_p[k]*U[k+1]+solver.G_p[k])/solver.beta_p[k];
    a = solver.D[k]*U_re[k-1]+solver.E_re[k]*U_re[k]-solver.E_im[k]*U_im[k]+solver.F[k]*U_re[k+1]+solver.G[k];
    b = solver.D[k]*U_im[k-1]+solver.E_re[k]*U_im[k]+solver.E_im[k]*U_re[k]+solver.F[k]*U_im[k+1]+solver.G[k];
    r2 = solver.beta_p_r2[k];
    solver.X_re[k] = (a*solver.beta_p_re[k]+b*solver.beta_p_im[k])/r2;
    solver.X_im[k] = (b*solver.beta_p_re[k]-a*solver.beta_p_im[k])/r2;
    kn = k;
    for (j=2; j<=solver.height; j++) {
        k += solver.full_width;
        //solver.X[k] = (solver.D_p[k]*U[k-1]+solver.E_p[k]*U[k]+solver.F_p[k]*U[k+1]+solver.G[k]-solver.A_p[k]*solver.X[kn])/solver.beta_p[k];
        a = solver.D[k]*U_re[k-1]+solver.E_re[k]*U_re[k]-solver.E_im[k]*U_im[k]+solver.F[k]*U_re[k+1]+solver.G[k]-solver.A[k]*solver.X_re[kn];
        b = solver.D[k]*U_im[k-1]+solver.E_re[k]*U_im[k]+solver.E_im[k]*U_re[k]+solver.F[k]*U_im[k+1]+solver.G[k]-solver.A[k]*solver.X_im[kn];
        r2 = solver.beta_p_r2[k];
        solver.X_re[k] = (a*solver.beta_p_re[k]+b*solver.beta_p_im[k])/r2;
        solver.X_im[k] = (b*solver.beta_p_re[k]-a*solver.beta_p_im[k])/r2;
        kn = k;
    }
    j = solver.height;
    k = IND(i,solver.height);
    kn = k;
    U1_re[k] = solver.X_re[k];
    U1_im[k] = solver.X_im[k];
    while (j>1) {
        k -= solver.full_width;
        j--;
        //U1[k] = solver.X[k]-solver.gamma_p[k]*U1[kn];
        U1_re[k] = solver.X_re[k]-(solver.gamma_p_re[k]*U1_re[kn]-solver.gamma_p_im[k]*U1_im[kn]);
        U1_im[k] = solver.X_im[k]-(solver.gamma_p_re[k]*U1_im[kn]+solver.gamma_p_im[k]*U1_re[kn]);
        kn = k;
    }
    return 0;
}


int _compute_proba(void) {
    float maxP = 0.0;
    float p;
    int k,c,i,j,k2,k0;
    float dz1,dz2;
    float r,g,b,m;
    for (k=0; k<solver.full_size; k++) {
        p = solver.U_re[k]*solver.U_re[k]+solver.U_im[k]*solver.U_im[k];
        solver.P[k] = p;
        if (p > maxP) maxP = p; 
    }
    
    if (solver.normalize) {
           for (k=0; k<solver.full_size; k++) solver.P[k] /= maxP;
    }
    if (solver.colormap) {
           k = 0; k2 = 0; k0 = 0;
           for (j=1; j<=solver.height; j++) {
                 for (i=1; i<=solver.width; i++) {
                     /*if ((solver.contours[k0]==1.0)&&(solver.cont==1)) {
                         solver.image[k2] = solver.cont_r;
                         k2++;
                         solver.image[k2] = solver.cont_g;
                         k2++;
                         solver.image[k2] = solver.cont_b;
                         k2++;
                         solver.image[k2] = 1.0;
                         k2++;
                     }
                     else {*/
                         c = solver.P[k]*255;
                         if (c>255) c=255;
                         r = solver.map_r[c]; g = solver.map_g[c]; b = solver.map_b[c];
                         if ((solver.contours[k0]==1.0)&&(solver.cont==1)) {
                            if (r > b) m = r; else m = b;
                            if (g > m) m=g;
                            m = 1.0-m;
                            r += m; g += m; b += m;
                         }
                         solver.image[k2] = r;
                         k2++;
                         solver.image[k2] = g;
                         k2++;
                         solver.image[k2] = b;
                         k2++;
                         solver.image[k2] = 1.0;
                         k2++;
                     //}
                     k++;
                     k0++;
                 }
                k+=2;
           }
    }
    if (solver.surface) {
        
           k = 0; k2 = 0;
           for (j=1; j<=solver.height; j++) {
                 for (i=1; i<=solver.width; i++) {
                         solver.sommets_z[k2++] = solver.P[k++]*solver.surface_height;
                         
                 }
                 k+=2;
           }
           k = 0; k2 = 0;
           for (j=1; j<=solver.height; j++) {
                for (i=1; i<=solver.width; i++) {
                       dz1 = solver.P[k+1]-solver.P[k];
                       dz2 = solver.P[k+solver.full_width]-solver.P[k];
                       solver.normales[k2++] = -solver.dy*dz1;
                       solver.normales[k2++] = -solver.dx*dz2;
                       solver.normales[k2++] = solver.dxdy;
                       k++;
                }
                k+=2;
           }
    }
    return 0;
}

int schrodinger2d_iterations(float ti, float tf) {
        int i,j;
        solver.t = ti;
        while (solver.t < tf) {
               for (j=1; j<=solver.height; j++) _row_iteration(j,solver.U_re,solver.U_im,solver.U1_re,solver.U1_im);
               for (i=1; i<=solver.width; i++) _col_iteration(i,solver.U1_re,solver.U1_im,solver.U_re,solver.U_im);
               solver.t += solver.dt;
        }
        _compute_proba();
        return 0;
}

void* _iterations_thread(void *arg) {
     threadArg* ta = (threadArg*)arg;
     int iStart = ta->iStart;
     int iEnd = ta->iEnd;
     int jStart = ta->jStart;
     int jEnd = ta->jEnd;
     int niter = ta->niter;
     int id = ta->id;
     int i,j;
     int iter;
     niter = ta->niter;
     for (iter=0; iter<niter; iter++) {
         for (j=jStart; j<jEnd; j++) _row_iteration(j,solver.U_re,solver.U_im,solver.U1_re,solver.U1_im);
         pthread_barrier_wait(&threadBarrier);
         for (i=iStart; i<iEnd; i++) _col_iteration(i,solver.U1_re,solver.U1_im,solver.U_re,solver.U_im);
         pthread_barrier_wait(&threadBarrier);
         if (id==0) solver.t += solver.dt;
     }
     pthread_exit(NULL);
}


void* _row_iteration_thread(void *arg) {
    threadArg* ta = (threadArg*)arg;
     int jStart = ta->jStart;
     int jEnd = ta->jEnd;
     int j;
     float *U_re = ta->U_re;
     float *U_im = ta->U_im;
     float *U1_re = ta->U1_re;
     float *U1_im = ta->U1_im;
     for (j=jStart; j<jEnd; j++) _row_iteration(j,U_re,U_im,U1_re,U1_im);
     pthread_exit(NULL);
} 

void * _col_iteration_thread(void *arg) {
     threadArg* ta = (threadArg*)arg;
     int iStart = ta->iStart;
     int iEnd = ta->iEnd;
     int i;
     float *U_re = ta->U_re;
     float *U_im = ta->U_im;
     float *U1_re = ta->U1_re;
     float *U1_im = ta->U1_im;
     for (i=iStart; i<iEnd; i++) _col_iteration(i,U_re,U_im,U1_re,U1_im);
     pthread_exit(NULL);
}



int schrodinger2d_iterations_2threads(float ti, float tf) {
    pthread_attr_t attr;
    pthread_t thread1;
    threadArg arg1,arg2;
    int err;
    int i,j;
    arg1.iStart = 1;
    arg1.iEnd = solver.width/2;
    arg2.iStart = arg1.iEnd;
    arg2.iEnd = solver.width+1;
    arg1.jStart = 1;
    arg1.jEnd = solver.height/2;
    arg2.jStart = arg1.jEnd;
    arg2.jEnd = solver.height+1;
    solver.t = ti;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    while (solver.t < tf) {
        arg1.U_re = solver.U_re;
        arg1.U_im = solver.U_im;
        arg1.U1_re = solver.U1_re;
        arg1.U1_im = solver.U1_im;
        err = pthread_create(&thread1,&attr,_row_iteration_thread,(void*)&arg1);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        for (j=arg2.jStart; j<arg2.jEnd; j++) _row_iteration(j,solver.U_re,solver.U_im,solver.U1_re,solver.U1_im);
        err = pthread_join(thread1,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        arg1.U_re = solver.U1_re;
        arg1.U_im = solver.U1_im;
        arg1.U1_re = solver.U_re;
        arg1.U1_im = solver.U_im;
        err = pthread_create(&thread1,&attr,_col_iteration_thread,(void*)&arg1);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        for (i=arg2.iStart; i<arg2.iEnd; i++) _col_iteration(i,solver.U1_re,solver.U1_im,solver.U_re,solver.U_im);
        err = pthread_join(thread1,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        solver.t += solver.dt;
    }
    pthread_attr_destroy(&attr);
    _compute_proba();
    return 0;
}

int schrodinger2d_iterations_4threads(float ti, float tf) {
    pthread_attr_t attr;
    pthread_t thread1,thread2,thread3;
    threadArg arg1,arg2,arg3,arg4;
    int err;
    int i,j;
    int part_i, part_j;
    part_i = solver.width/4;
    part_j = solver.height/4;
    arg1.iStart = 1;
    arg1.iEnd = part_i;
    arg1.jStart = 1;
    arg1.jEnd = part_j;
    arg2.iStart = arg1.iEnd;
    arg2.iEnd = 2*part_i;
    arg2.jStart = arg1.jEnd;
    arg2.jEnd = 2*part_j;
    arg3.iStart = arg2.iEnd;
    arg3.iEnd = 3*part_i;
    arg3.jStart = arg2.jEnd;
    arg3.jEnd = 3*part_j;
    arg4.iStart = arg3.iEnd;
    arg4.iEnd = solver.width+1;
    arg4.jStart = arg3.jEnd;
    arg4.jEnd = solver.height+1;
    solver.t = ti;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    while (solver.t < tf) {
        arg1.U_re = solver.U_re; arg1.U_im = solver.U_im; arg1.U1_re = solver.U1_re; arg1.U1_im = solver.U1_im;
        arg2.U_re = solver.U_re; arg2.U_im = solver.U_im; arg2.U1_re = solver.U1_re; arg2.U1_im = solver.U1_im;
        arg3.U_re = solver.U_re; arg3.U_im = solver.U_im; arg3.U1_re = solver.U1_re; arg3.U1_im = solver.U1_im;
        err = pthread_create(&thread1,&attr,_row_iteration_thread,(void*)&arg1);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        err = pthread_create(&thread2,&attr,_row_iteration_thread,(void*)&arg2);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        err = pthread_create(&thread3,&attr,_row_iteration_thread,(void*)&arg3);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        for (j=arg4.jStart; j<arg4.jEnd; j++) _row_iteration(j,solver.U_re,solver.U_im,solver.U1_re,solver.U1_im);
        err = pthread_join(thread1,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        err = pthread_join(thread2,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        err = pthread_join(thread3,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        arg1.U_re = solver.U1_re; arg1.U_im = solver.U1_im; arg1.U1_re = solver.U_re; arg1.U1_im = solver.U_im;
        arg2.U_re = solver.U1_re; arg2.U_im = solver.U1_im; arg2.U1_re = solver.U_re; arg2.U1_im = solver.U_im;
        arg3.U_re = solver.U1_re; arg3.U_im = solver.U1_im; arg3.U1_re = solver.U_re; arg3.U1_im = solver.U_im;
        err = pthread_create(&thread1,&attr,_col_iteration_thread,(void*)&arg1);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        err = pthread_create(&thread2,&attr,_col_iteration_thread,(void*)&arg2);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        err = pthread_create(&thread3,&attr,_col_iteration_thread,(void*)&arg3);
        if (err) return SCHRODINGER2D_ERROR_THREAD_START;
        for (i=arg4.iStart; i<arg4.iEnd; i++) _col_iteration(i,solver.U1_re,solver.U1_im,solver.U_re,solver.U_im);
        err = pthread_join(thread1,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        err = pthread_join(thread2,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        err = pthread_join(thread3,NULL);
        if (err) return SCHRODINGER2D_ERROR_THREAD_JOIN;
        solver.t += solver.dt;
    }
    pthread_attr_destroy(&attr);
    _compute_proba();
    return 0;
}

int _paquet(float x0, float y0, float k0, float sigma0) {
    int i,j,k;
    float x,y;
    float r2;
    float a,e;
    a = 1.0/(4*sigma0*sigma0);
    for (i=2; i<solver.width; i++) {
        for (j=2; j<solver.height; j++) {
              x = i*solver.dx;
              y = j*solver.dy;
              r2 = (x-x0)*(x-x0)+(y-y0)*(y-y0);
              e = exp(-r2*a);
              if (e>0.0) {
                  k = IND(i,j);
                  solver.U_re[k] = e*cos(k0*x);
                  solver.U_im[k] = e*sin(k0*x);
              }
        }
    }
    return 0;   
}

int schrodinger2d_paquet(float x0, float y0, float k0, float sigma0) {
    switch (solver.type) {
    case TYPE_CARTESIAN: _paquet(x0,y0,k0,sigma0);
    }
    return 0;
}



int schrodinger2d_get_psi(float *psi_re, float *psi_im) {
    int k,m,j;
    int size;
    size = solver.width*sizeof(float);
    k = IND(1,1);
    m = 0;
    for (j=1; j<=solver.height; j++) {
        memcpy(psi_re+m,solver.U_re+k,size);
        memcpy(psi_im+m,solver.U_im+k,size);
        m += solver.width;
        k += solver.full_width;
    }
    return 0;
}



int schrodinger2d_get_proba(float *proba) {
    int k,m,j;
    int size;
    size = solver.width*sizeof(float);
    k = IND(1,1);
    m = 0;
    for (j=1; j<=solver.height; j++) {
        memcpy(proba+m,solver.P+k,size);
        m += solver.width;
        k += solver.full_width;
    }
    return 0;
}



int schrodinger2d_get_proba_colors(float* image) {
    memcpy(image,solver.image,solver.size*4*sizeof(float));
    return 0;
}



int schrodinger2d_get_contours(float *data) {
    memcpy(data,solver.contours,solver.size*sizeof(float));
    return 0;
}

int schrodinger2d_get_size(int *width, int *height) {
    *width = solver.width;
    *height = solver.height;
    return 0;
}


/*----------------------------- OpenGL -----------------------------------------------------------------------*/



void _gl_display_image(void) {
    glBindTexture(GL_TEXTURE_2D,solver.tex_rgb);
    glPixelStoref(GL_UNPACK_ALIGNMENT,1);
    glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,solver.width,solver.height,0,GL_RGBA,GL_FLOAT,solver.image);
    glClear(GL_COLOR_BUFFER_BIT);
    draw_image_rgb(solver.gl_program_image,solver.tex_rgb);
    
    glutSwapBuffers();
    
}

void _gl_display2(void) {
    glClear(GL_COLOR_BUFFER_BIT);
    draw_image_rgb(solver.gl_program_image,solver.tex_rgb);
    glutSwapBuffers();
}



void _gl_buffers_data(void) {
    glBindBuffer(GL_ARRAY_BUFFER,solver.buffer_sommets_z);
    glBufferData(GL_ARRAY_BUFFER,sizeof(float)*solver.size,solver.sommets_z,GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER,solver.buffer_normales);
    glBufferData(GL_ARRAY_BUFFER,sizeof(float)*3*solver.size,solver.normales,GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER,solver.buffer_colors);
    glBufferData(GL_ARRAY_BUFFER,sizeof(float)*4*solver.size,solver.image,GL_DYNAMIC_DRAW);
}


void _gl_display_surface(void) {
    
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     transfo_copy(solver.transfo_PMV,solver.transfo_P);
     transfo_compose(solver.transfo_PMV,solver.transfo_MV);
     _gl_buffers_data();
     transfo_get_normal_matrix(solver.transfo_MV,solver.Nmatrix);
     
     glBindTexture(GL_TEXTURE_2D,solver.tex_rgb);
    glPixelStoref(GL_UNPACK_ALIGNMENT,1);
    glTexImage2D(GL_TEXTURE_2D,0,GL_RED,solver.width,solver.height,0,GL_RED,GL_FLOAT,solver.contours);
    draw_pixmap_modelview(solver.gl_program_pixmap_modelview,solver.tex_rgb,solver.x0,-solver.x0,solver.y0,-solver.y0,solver.transfo_PMV->M);
    
     draw_surface(solver.gl_program_surface,solver.nb_sommets,solver.buffer_indices,solver.buffer_sommets_xy,solver.buffer_sommets_z,solver.buffer_normales,solver.buffer_colors,solver.transfo_PMV->M,solver.transfo_MV->M,solver.Nmatrix,solver.LightPosition,solver.Kdiffuse,solver.LightIntensity,solver.Kambient);
     
     
     
     glutSwapBuffers();
     
}

void _gl_reshape(int width, int height) {
    glViewport(0,0,width,height);
    
}

void _gl_fbo_render_surface(void) {
    GLuint fbo,rbo,rbo_d;
    glGenFramebuffers(1,&fbo);
    glBindFramebuffer(GL_FRAMEBUFFER,fbo);
    glGenRenderbuffers(1,&rbo);
    glBindRenderbuffer(GL_RENDERBUFFER,rbo);
    glRenderbufferStorage(GL_RENDERBUFFER,GL_RGBA,solver.win_width,solver.win_height);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,GL_RENDERBUFFER,rbo);
    glGenRenderbuffers(1,&rbo_d);
    glBindRenderbuffer(GL_RENDERBUFFER,rbo_d);
    glRenderbufferStorage(GL_RENDERBUFFER,GL_DEPTH_COMPONENT,solver.win_width,solver.win_height);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER,GL_DEPTH_ATTACHMENT,GL_RENDERBUFFER,rbo_d);
    glDrawBuffer(GL_FRONT);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     transfo_copy(solver.transfo_PMV,solver.transfo_P);
     transfo_compose(solver.transfo_PMV,solver.transfo_MV);
     _gl_buffers_data();
     transfo_get_normal_matrix(solver.transfo_MV,solver.Nmatrix);
     glViewport(0,0,solver.win_width,solver.win_height);
     draw_surface(solver.gl_program_surface,solver.nb_sommets,solver.buffer_indices,solver.buffer_sommets_xy,solver.buffer_sommets_z,solver.buffer_normales,solver.buffer_colors,solver.transfo_PMV->M,solver.transfo_MV->M,solver.Nmatrix,solver.LightPosition,solver.Kdiffuse,solver.LightIntensity,solver.Kambient);
    
     glBindTexture(GL_TEXTURE_2D,solver.tex_rgb);
    glPixelStoref(GL_UNPACK_ALIGNMENT,1);
    glTexImage2D(GL_TEXTURE_2D,0,GL_RED,solver.width,solver.height,0,GL_RED,GL_FLOAT,solver.contours);
    draw_pixmap_modelview(solver.gl_program_pixmap_modelview,solver.tex_rgb,solver.x0,-solver.x0,solver.y0,-solver.y0,solver.transfo_PMV->M);
     
     glReadPixels(0,0,solver.win_width,solver.win_height,GL_RGB,GL_UNSIGNED_BYTE,solver.render_img);
    
     glBindRenderbuffer(GL_RENDERBUFFER,0);
    glBindFramebuffer(GL_FRAMEBUFFER,0);
}

void _gl_fbo_render_image(void) {
    GLuint fbo,rbo;
    glGenFramebuffers(1,&fbo);
    glBindFramebuffer(GL_FRAMEBUFFER,fbo);
    glGenRenderbuffers(1,&rbo);
    glBindRenderbuffer(GL_RENDERBUFFER,rbo);
    glRenderbufferStorage(GL_RENDERBUFFER,GL_RGBA,solver.win_width,solver.win_height);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,GL_RENDERBUFFER,rbo);
    
    glDrawBuffer(GL_FRONT);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glBindTexture(GL_TEXTURE_2D,solver.tex_rgb);
    glPixelStoref(GL_UNPACK_ALIGNMENT,1);
    glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,solver.width,solver.height,0,GL_RGBA,GL_FLOAT,solver.image);
    glClear(GL_COLOR_BUFFER_BIT);
    draw_image_rgb(solver.gl_program_image,solver.tex_rgb);
     glReadPixels(0,0,solver.win_width,solver.win_height,GL_RGB,GL_UNSIGNED_BYTE,solver.render_img);
    
     glBindRenderbuffer(GL_RENDERBUFFER,0);
    glBindFramebuffer(GL_FRAMEBUFFER,0);  
}

void _iteration_loop(void) {
     int i,j;
     if (solver.t < solver.tf) {
           for (j=1; j<=solver.height; j++) _row_iteration(j,solver.U_re,solver.U_im,solver.U1_re,solver.U1_im);
           for (i=1; i<=solver.width; i++) _col_iteration(i,solver.U1_re,solver.U1_im,solver.U_re,solver.U_im);
           solver.t += solver.dt;
     }
     _compute_proba();
     glutPostRedisplay();
}



void _opencl_iteration_loop(void) {
    if (solver.t < solver.tf) {
         schrodinger2d_opencl_iterations(solver.t,solver.t+solver.dt);
         solver.t += solver.dt;
    }
    glutPostRedisplay();
}

void _init_glut(void) {
    char *name="plot";
    char *argv[1];
    int argc = 1;
    GLuint err;
    argv[0] = name;
    if (solver.glut_init==1) return;
    solver.glut_init = 1;
    glutInit(&argc,argv);
    glutInitContextVersion(3,0);
    glutInitContextProfile(GLUT_CORE_PROFILE);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    
    glutInitWindowSize(solver.win_width,solver.win_height);
    glutCreateWindow("Schrodinger");
    
    
    err = glewInit();
    if (err != GLEW_OK) { 
        printf("GLEW ERROR : %s\n",glewGetErrorString(err));  
    }
}


int _init_gl(int rendering) {
    
    if (rendering==RENDERING_IMG) {
        glutDisplayFunc(_gl_display_image);
        glDisable(GL_DEPTH_TEST);
        solver.surface=0;
    }
    else {
        glutDisplayFunc(_gl_display_surface);
        glEnable(GL_DEPTH_TEST);
        solver.surface = 1;
    }
    glViewport(0,0,solver.win_width,solver.win_height);
    glutReshapeFunc(_gl_reshape);
    glClearColor(0.0f,0.0f,0.0f,1.0f);
    
    solver.gl_program_image = get_gl_program_image_rgb();
    solver.gl_program_surface = get_gl_program_surface();
    solver.gl_program_pixmap_modelview = get_gl_program_pixmap_modelview();
    
    glGenTextures(1,&solver.tex_rgb);
    //glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D,solver.tex_rgb);
    glPixelStoref(GL_UNPACK_ALIGNMENT,1);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,solver.width,solver.height,0,GL_RGBA,GL_FLOAT,NULL);
    glBindTexture(GL_TEXTURE_2D,0);
    
    if (solver.surface) {
        glGenBuffers(1,&solver.buffer_indices);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,solver.buffer_indices);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER,sizeof(uint32_t)*solver.nb_sommets,solver.indices,GL_STATIC_DRAW);
        glGenBuffers(1,&solver.buffer_sommets_xy);
        glBindBuffer(GL_ARRAY_BUFFER,solver.buffer_sommets_xy);
        glBufferData(GL_ARRAY_BUFFER,solver.size*2*sizeof(float),solver.sommets_xy,GL_STATIC_DRAW);
        glGenBuffers(1,&solver.buffer_sommets_z);
        glGenBuffers(1,&solver.buffer_normales);
        glGenBuffers(1,&solver.buffer_colors);
    }
    solver.LightPosition[0] = 0;
    solver.LightPosition[1] = -1;
    solver.LightPosition[2] = 1;
    solver.LightPosition[3] = 0;
    solver.Kdiffuse[0] = 0.5;
    solver.Kdiffuse[1] = 0.5;
    solver.Kdiffuse[2] = 0.5;
    solver.LightIntensity[0] = 1.0;
    solver.LightIntensity[1] = 1.0;
    solver.LightIntensity[2] = 1.0;
    solver.Kambient[0] = 0.5;
    solver.Kambient[1] = 0.5;
    solver.Kambient[2] = 0.5;
    
    
    return 0;
}

void _gl_close(void) {
    glDeleteBuffers(1,&solver.buffer_indices);
    glDeleteBuffers(1,&solver.buffer_sommets_xy);
    glDeleteBuffers(1,&solver.buffer_sommets_z);
    glDeleteBuffers(1,&solver.buffer_normales);
    glDeleteBuffers(1,&solver.buffer_colors);   
}




int schrodinger2d_start_rendering(float ti, float tf, int opencl, int width, int height, float surface_height, int rendering, float angle_x, float angle_z) {

    solver.surface_height = surface_height; 
    solver.win_width = width;
    solver.win_height = height;
    if (rendering==RENDERING_IMG) solver.cont=1;
    else solver.cont = 0;
    _init_glut();
    _init_gl(rendering);
    projection_ortho(solver.transfo_P,1,1,-1,1);
    
    transfo_rotate_x(solver.transfo_MV,-angle_x);
    transfo_rotate_z(solver.transfo_MV,-angle_z);
    
    if (opencl) glutIdleFunc(_opencl_iteration_loop);
    else glutIdleFunc(_iteration_loop);
    solver.t = ti;
    solver.tf = tf;
    glutMainLoop();
    _gl_close();
    solver.cont = 1;
    return 0;
}

int schrodinger2d_start_gl() {
    solver.win_width = 0;
    solver.win_height = 0;
    _init_glut();
    return 0;
}

int schrodinger2d_offscreen_rendering(float ti, float tf, int opencl,int width, int height, float surface_height, int rendering, float angle_x, float angle_z, void *data) {
    
    solver.render_img = (uint8_t*) data;
    solver.surface_height = surface_height;
    if (rendering==RENDERING_IMG) solver.cont=1;
    else solver.cont = 0;
    if (rendering==RENDERING_SURFACE) solver.surface = 1;
    if (opencl) {
        schrodinger2d_opencl_iterations(ti,tf);
    }
    else {
        schrodinger2d_iterations_2threads(ti,tf);
    }
    solver.win_width = 1;
    solver.win_height = 1;
    _init_glut();
    solver.win_width = width;
    solver.win_height = height;
    _init_gl(rendering);
    projection_ortho(solver.transfo_P,1,1,-1,1);
    
    transfo_init(solver.transfo_MV);
    transfo_rotate_x(solver.transfo_MV,-angle_x);
    transfo_rotate_z(solver.transfo_MV,-angle_z);
    
    
    if (rendering==RENDERING_SURFACE) _gl_fbo_render_surface();
    if (rendering==RENDERING_IMG) _gl_fbo_render_image();
    _gl_close();
    return 0;
}
/*------------------------------------------------- openCL -----------------------------------------------*/
#ifdef OPENCL
cl_uint numPlatforms;
cl_device_type deviceType;
cl_platform_id platform;
cl_device_id device;
size_t maxWorkGroupSize;
cl_uint maxWorkItemDimensions;
size_t maxWorkItemSizes[3];
cl_context context;
cl_program OpenCLProgram;
cl_command_queue commandQueue;
cl_kernel row_iteration, col_iteration, proba, norm_proba,map_colors_img,gen_surface;



const char* programSource[] = {
"__kernel void row_iteration(__global float *U_re,__global float *U_im, __global float *U1_re, __global float *U1_im, __global float *A, __global float *D, __global float *E_re, __global float *E_im, __global float *F, __global float *G, __global float *beta_re, __global float *beta_im, __global float *beta_r2, __global float *gamma_re, __global float *gamma_im,short nx, short ny, short Nx, short Ny) {",
            "float x_x_re[1024];",
            "float x_x_im[1024];",
            "int j = get_global_id(0);",
            "int i,k;",
            "float a,b,r2;",
            "float L_re,L_im;",
            "i = 1;",            
            "k = i+j*nx;",
            "a = D[k]*U_re[k-nx]+E_re[k]*U_re[k]-E_im[k]*U_im[k]+F[k]*U_re[k+nx]+G[k];",
            "b = D[k]*U_im[k-nx]+E_re[k]*U_im[k]+E_im[k]*U_re[k]+F[k]*U_im[k+nx]+G[k];",
            "r2 = beta_r2[k];",
            "L_re = x_x_re[i] = (a*beta_re[k]+b*beta_im[k])/r2;",
            "L_im = x_x_im[i] = (b*beta_re[k]-a*beta_im[k])/r2;",
            "while (i<Nx) {",
                "k += 1;",
                "i++;",
                "a = D[k]*U_re[k-nx]+E_re[k]*U_re[k]-E_im[k]*U_im[k]+F[k]*U_re[k+nx]+G[k]-A[k]*L_re;",
                "b = D[k]*U_im[k-nx]+E_re[k]*U_im[k]+E_im[k]*U_re[k]+F[k]*U_im[k+nx]+G[k]-A[k]*L_im;",
                "r2 = beta_r2[k];",
                "L_re = x_x_re[i] = (a*beta_re[k]+b*beta_im[k])/r2;",
                "L_im = x_x_im[i] = (b*beta_re[k]-a*beta_im[k])/r2;",
            "}",
            "i = Nx;",
            "k = i+j*nx;",
            "L_re = U1_re[k] = x_x_re[i];",
            "L_im = U1_im[k] = x_x_im[i];",
            "while (i>1) {",
                "k-=1;",
                "i--;",
                "U1_re[k] = x_x_re[i]-(gamma_re[k]*L_re-gamma_im[k]*L_im);",
                "U1_im[k] = x_x_im[i]-(gamma_re[k]*L_im+gamma_im[k]*L_re);",
                "L_re = U1_re[k];",
                "L_im = U1_im[k];",
            "}",
        "}",
"__kernel void col_iteration(__global float *U_re,__global float *U_im, __global float *U1_re, __global float *U1_im, __global float *A, __global float *D, __global float *E_re, __global float *E_im, __global float *F, __global float *G, __global float *beta_p_re, __global float *beta_p_im, __global float *beta_p_r2, __global float *gamma_p_re, __global float *gamma_p_im,short nx, short ny, short Nx, short Ny) {",
            "float x_y[1024];",
            "float x_y[1024];",
            "int i = get_global_id(0);",
            "int j,k;",
            "float a,b,r2;",
            "float L_re,L_im;",
            "j = 1;",
            "k = i+j*nx;",
            "a = D[k]*U_re[k-1]+E_re[k]*U_re[k]-E_im[k]*U_im[k]+F[k]*U_re[k+1]+G[k];",
            "b = D[k]*U_im[k-1]+E_re[k]*U_im[k]+E_im[k]*U_re[k]+F[k]*U_im[k+1]+G[k];",
            "r2 = beta_p_r2[k];",
            "L_re = x_y_re[j] = (a*beta_p_re[k]+b*beta_p_im[k])/r2;",
            "L_im = x_y_im[j] = (b*beta_p_re[k]-a*beta_p_im[k])/r2;",
            "while (j<Ny) {",
                "k += nx;",
                "j++;",
                "a = D[k]*U_re[k-1]+E_re[k]*U_re[k]-E_im[k]*U_im[k]+F[k]*U_re[k+1]+G[k]-A[k]*L_re;",
                "b = D[k]*U_im[k-1]+E_re[k]*U_im[k]+E_im[k]*U_re[k]+F[k]*U_im[k+1]+G[k]-A[k]*L_im;",
                "r2 = beta_p_r2[k];",
                "L_re=x_y_re[j] = (a*beta_p_re[k]+b*beta_p_im[k])/r2;",
                "L_im=x_y_im[j] = (b*beta_p_re[k]-a*beta_p_im[k])/r2;",
            "}",
            "j = Ny;",
            "k = i+j*nx;",
            "L_re = U1_re[k] = x_y_re[j];",
            "L_im = U1_im[k] = x_y_im[j];",
            "while (j>1) {",
                "k -= nx;",
                "j--;",
                "U1_re[k] = x_y_re[j]-(gamma_p_re[k]*L_re-gamma_p_im[k]*L_im);",
                "U1_im[k] = x_y_im[j]-(gamma_p_re[k]*L_im+gamma_p_im[k]*L_re);",
                "L_re = U1_re[k];",
                "L_im = U1_im[k];",
            "}",
        "}",
"__kernel void proba(__global float *U_re, __global float *U_im, __global float *P, __global float *Pmax, short nx, short ny, short ngx, short ngy, __local float *row_max) {",
            "int i = get_global_id(0);",
            "int j = get_global_id(1);",
            "int k = i+j*nx;",
            "float re = U_re[k];",
            "float im = U_im[k];",
            "P[k] = re*re+im*im;",
            "barrier(CLK_GLOBAL_MEM_FENCE);",
            "int size_i0 = get_local_size(0);",
            "int size_j0 = get_local_size(1);",
            "int i0 = get_local_id(0);",
            "int j0 = get_local_id(1);",
            "k = nx*(get_group_id(1)*size_j0 + j0 + 1) + get_group_id(0)*size_i0 +1;",
            "float max = 0.0;",
            "float p;",
            "if (i0==0) {",
                "for (i0=0; i0 < size_i0; i0++) {",
                    "p=P[k];",
                    "if (p > max) max = p;",
                    "k++;",
                "}",
                "row_max[j0] = max;",
                "barrier(CLK_LOCAL_MEM_FENCE);",
                "if (j0==0) {",
                    "for (j0=1; j0<size_j0; j0++) {",
                            "if (row_max[j0] > max) max = row_max[j0];",
                    "}",
                    "Pmax[get_group_id(0)+ngx*get_group_id(1)] = max;",
                "}",
            "}",
        "}",
"__kernel void norm_proba(__global float *P, float pmax, short nx, short ny){",
            "int i = get_global_id(0);",
            "int j = get_global_id(1);",
            "int k = i+j*nx;",
            "P[k] /= pmax;",
        "}",
"__constant sampler_t sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP | CLK_FILTER_NEAREST;",
"__kernel void map_colors_img(__global float *P,write_only image2d_t image, __global float *map_r, __global float *map_g, global float *map_b, __global float *contours, short nx, short ny, short Nx, short Ny, float cont_r, float cont_b, float cont_g, short cont){",
        "int i = get_global_id(0);",
        "int j = get_global_id(1);",
        "int k = i+1+(j+1)*nx;",
        "int k0 = i+j*Nx;",
        "float c = contours[k0];",
        "float4 pixel;",
        "float r,g,b,m;",
            "int cp = P[k]*255;",
            "if (cp>255) cp=255;",
            "r = map_r[cp]; g = map_g[cp]; b = map_b[cp];",
            "if ((c > 0.0)&&(cont==1)) {m = 1.0-max(r,max(g,b)); r+=m; g+=m; b+=m;}",
            "pixel = (float4)(r,g,b,1.0);",
        "int2 coord = (int2)(i,j);",
        "write_imagef(image,coord,pixel);",
    "}",
"__kernel void gen_surface(__global float *P,__global float *sommets_z, __global float *normales, float surface_height, float dx, float dy, float dxdy, short nx, short ny, short Nx, short Ny) {",
        "int i = get_global_id(0);",
        "int j = get_global_id(1);",
        "int k = i+1+(j+1)*nx;",
        "int k2 = i+j*Nx;",
        "int k3 = k2*3;",
        "float p = P[k];",
        "sommets_z[k2] = p*surface_height;",
        "float dz1 = P[k+1]-p;",
        "float dz2 = P[k+nx]-p;",
        "normales[k3] = -dy*dz1;",
        "normales[k3+1] = -dx*dz2;",
        "normales[k3+2] = dxdy;",
"}"
};
#define SOURCE_NB_LINES 140
#define OPENCL_ERROR(e,msg) if (e<0) {printf("Erreur openCL : %s : %d\n",msg,e);return e;}

int schrodinger2d_opencl_platforms() {
    cl_int err;
    int p,k;
    char *name;
    size_t s;
    cl_platform_id* platforms;
    cl_uint numDev;
    cl_device_id* devices;
    err = clGetPlatformIDs(0,NULL,&numPlatforms);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    printf("Nombre de platformes openCL : %d\n",numPlatforms);
    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id)*numPlatforms);
    err = clGetPlatformIDs(numPlatforms,platforms,NULL);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    for (p=0; p<numPlatforms; p++) {
        err = clGetPlatformInfo(platforms[p],CL_PLATFORM_NAME,0,NULL,&s);
        OPENCL_ERROR(err,"clGetPlatformInfo")
        name = (char*)malloc(sizeof(char)*s);
        err = clGetPlatformInfo(platforms[p],CL_PLATFORM_NAME,s,name,NULL);
        OPENCL_ERROR(err,"clGetPlatformInfo")
        printf("Plateforme OpenCL %d : %s\n",p,name);
        free(name);
        //recherche des périphériques
        err = clGetDeviceIDs(platforms[p],CL_DEVICE_TYPE_ALL,0,NULL,&numDev);
        OPENCL_ERROR(err,"clGetDeviceIDs")
        printf("\t\tNombre de peripheriques : %d\n",numDev);
        devices = (cl_device_id*)malloc(sizeof(cl_device_id)*numDev);
        err = clGetDeviceIDs(platforms[p],CL_DEVICE_TYPE_ALL,numDev,devices,NULL);
        OPENCL_ERROR(err,"clGetDeviceIDs")
        for (k=0; k<numDev; k++) {
            err = clGetDeviceInfo(devices[k],CL_DEVICE_NAME,0,NULL,&s);
            OPENCL_ERROR(err,"clGetDeviceInfo")
            name = (char*)malloc(sizeof(char)*s);
            err = clGetDeviceInfo(devices[k],CL_DEVICE_NAME,s,name,NULL);
            OPENCL_ERROR(err,"clGetDeviceInfo")
            printf("\t\tPeripherique de calcul %d : %s\n",k,name);
            free(name); 
        }
        free(devices);
        printf("\n");
    }
    free(platforms);
    return 0;
}

int schrodinger2d_set_opencl_platform_device(int p, int d) {
    cl_int err;
    char *name;
    cl_platform_id* platforms;
    size_t s;
    cl_device_id* devices;
    cl_uint numDev;
    err = clGetPlatformIDs(0,NULL,&numPlatforms);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id)*numPlatforms);
    err = clGetPlatformIDs(numPlatforms,platforms,NULL);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    if ((p<0)||(p>=numPlatforms)) return SCHRODINGER2D_BAD_PLATFORM_NUMBER;
    platform = platforms[p];
    err = clGetPlatformInfo(platform,CL_PLATFORM_NAME,0,NULL,&s);
    OPENCL_ERROR(err,"clGetPlatformInfo")
    name = (char*)malloc(sizeof(char)*s);
    err = clGetPlatformInfo(platform,CL_PLATFORM_NAME,s,name,NULL);
    OPENCL_ERROR(err,"clGetPlatformInfo")
    printf("Plateforme OpenCL choisie : %s\n",name);
    free(name);
    free(platforms);
    err = clGetDeviceIDs(platform,CL_DEVICE_TYPE_ALL,0,NULL,&numDev);
    OPENCL_ERROR(err,"clGetDeviceIDs")
    devices = (cl_device_id*)malloc(sizeof(cl_device_id)*numDev);
    err = clGetDeviceIDs(platform,CL_DEVICE_TYPE_ALL,numDev,devices,NULL);
    if ((d<0)||(d>=numDev)) return SCHRODINGER2D_BAD_DEVICE_NUMBER;
    device = devices[d];
    free(devices);
    err = clGetDeviceInfo(device,CL_DEVICE_NAME,0,NULL,&s);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    name = (char*)malloc(sizeof(char)*s);
    err = clGetDeviceInfo(device,CL_DEVICE_NAME,s,name,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    printf("Peripherique de calcul choisi : %s",name);
    free(name);
    err = clGetDeviceInfo(device,CL_DEVICE_TYPE,sizeof(cl_device_type),&deviceType,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    if (deviceType==CL_DEVICE_TYPE_GPU) printf("\t (GPU)");
    else if (deviceType==CL_DEVICE_TYPE_CPU) printf("\t (CPU)");
    printf("\n");
    err = clGetDeviceInfo(device,CL_DEVICE_EXTENSIONS,0,NULL,&s);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    name = (char*)malloc(sizeof(char)*s);
    err = clGetDeviceInfo(device,CL_DEVICE_EXTENSIONS,s,name,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    //printf("Extensions : %s",name);
    free(name);
    return 0;
}

int schrodinger2d_opencl_init() {
    int err;
    size_t wg_size;
    size_t log_size;
    char *program_log;
    char decl_x_x_re[30];
    char decl_x_x_im[30];
    char decl_x_y_re[30];
    char decl_x_y_im[30];
    //int d;
    cl_context_properties context_props[7];
    
    if (platform==0) {
        err = schrodinger2d_set_opencl_platform_device(0,0);
        if (err<0) return err;
    }
    
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,sizeof(maxWorkItemDimensions),&maxWorkItemDimensions,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    //printf("Max work group item dimensions : %d\n",maxWorkItemDimensions);
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_SIZES,sizeof(maxWorkItemSizes),maxWorkItemSizes,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    
    //for (d=0; d<maxWorkItemDimensions; d++) printf("Max work item in workgroup for dimension %d : %d\n",d+1,maxWorkItemSizes[d]);
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_GROUP_SIZE,sizeof(maxWorkGroupSize),&maxWorkGroupSize,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    //printf("Max work group size : %d\n",maxWorkGroupSize);
     context_props[0] = CL_CONTEXT_PLATFORM;
    context_props[1] = (cl_context_properties)platform;
    /*context_props[2] = CL_GLX_DISPLAY_KHR;
    context_props[3] = (cl_context_properties) glXGetCurrentDisplay();
    context_props[4] = CL_GL_CONTEXT_KHR;
    context_props[5] = (cl_context_properties) glXGetCurrentContext();*/
    context_props[2] = 0;
    context = clCreateContextFromType(context_props,deviceType, NULL, NULL, &err);
    
    sprintf(decl_x_x_re,"float x_x_re[%d];",solver.full_width);
    sprintf(decl_x_x_im,"float x_x_im[%d];",solver.full_width);
    sprintf(decl_x_y_re,"float x_y_re[%d];",solver.full_height);
    sprintf(decl_x_y_im,"float x_y_im[%d];",solver.full_height);
    
    programSource[1] = decl_x_x_re;
    programSource[2] = decl_x_x_im;
    programSource[37] = decl_x_y_re;
    programSource[38] = decl_x_y_im;
    OpenCLProgram = clCreateProgramWithSource(context,SOURCE_NB_LINES,programSource,NULL,NULL);
    OPENCL_ERROR(err,"clCreateProgramWithSource")
    err = clBuildProgram(OpenCLProgram, 0, NULL, NULL, NULL, NULL);
    if (err<0) {
        printf("clBuildProgram error :\n");
        clGetProgramBuildInfo(OpenCLProgram,device,CL_PROGRAM_BUILD_LOG,0,NULL,&log_size);
        program_log = (char*) malloc(log_size+1);
        program_log[log_size] = '\0';
        clGetProgramBuildInfo(OpenCLProgram,device,CL_PROGRAM_BUILD_LOG,log_size+1,program_log,NULL);
        printf("%s\n",program_log);
        free(program_log);
    }
    
    OPENCL_ERROR(err,"clBuildProgram")
    commandQueue = clCreateCommandQueue(context, device, 0, &err);
    OPENCL_ERROR(err,"clCreateCommandQueue") 
    row_iteration = clCreateKernel(OpenCLProgram, "row_iteration",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    col_iteration = clCreateKernel(OpenCLProgram, "col_iteration",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    proba = clCreateKernel(OpenCLProgram, "proba",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    norm_proba = clCreateKernel(OpenCLProgram, "norm_proba",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    map_colors_img = clCreateKernel(OpenCLProgram, "map_colors_img",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    gen_surface = clCreateKernel(OpenCLProgram, "gen_surface",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    err = clGetKernelWorkGroupInfo(row_iteration,device,CL_KERNEL_WORK_GROUP_SIZE,sizeof(wg_size),&wg_size,NULL);
    OPENCL_ERROR(err,"clGetKernelWorkGroupSize")
    printf("max work group size : %d\n",wg_size);
    return 0;
}

int schrodinger2d_opencl_create_memory() {
    int err;
    cl_image_format format;
    
    solver.gpu_U_re = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR, sizeof(float)*solver.full_size,solver.U_re,&err);
    OPENCL_ERROR(err,"clCreateBuffer U_re")
    solver.gpu_U_im = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR, sizeof(float)*solver.full_size,solver.U_im,&err);
    OPENCL_ERROR(err,"clCreateBuffer U_im")
    solver.gpu_U1_re = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR, sizeof(float)*solver.full_size,solver.U1_re,&err);
    OPENCL_ERROR(err,"clCreateBuffer U1_re")
    solver.gpu_U1_im = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR, sizeof(float)*solver.full_size,solver.U1_im,&err);
    OPENCL_ERROR(err,"clCreateBuffer U1_im")
    solver.gpu_P = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR, sizeof(float)*solver.full_size,solver.P,&err);
    OPENCL_ERROR(err,"clCreateBuffer P")
    solver.gpu_Pmax = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR, sizeof(float)*solver.Pmax_size_x*solver.Pmax_size_y,solver.Pmax,&err);
    OPENCL_ERROR(err,"clCreateBuffer Pmax")
    solver.gpu_A = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.A,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_D = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.D,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_E_re = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.E_re,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_E_im = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.E_im,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_F = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.F,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_G = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.G,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_beta_re = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.beta_re,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_beta_im = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.beta_im,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_beta_r2 = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.beta_r2,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_gamma_re = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.gamma_re,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_gamma_im = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.gamma_im,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_beta_p_re = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.beta_p_re,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_beta_p_im = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.beta_p_im,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_beta_p_r2 = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.beta_p_r2,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_gamma_p_re = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.gamma_p_re,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_gamma_p_im = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*solver.full_size,solver.gamma_p_im,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_map_r = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*256,solver.map_r,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_map_g = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*256,solver.map_g,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_map_b = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*256,solver.map_b,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_contours = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR, sizeof(float)*solver.size,solver.contours,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    format.image_channel_order = CL_RGBA;
    format.image_channel_data_type = CL_FLOAT;
    solver.gpu_image = clCreateImage2D(context,CL_MEM_WRITE_ONLY,&format,solver.width,solver.height,0,NULL,&err);
    OPENCL_ERROR(err,"clCreateImage")
    solver.gpu_sommets_z = clCreateBuffer(context,CL_MEM_WRITE_ONLY,sizeof(float)*solver.size,NULL,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    solver.gpu_normales = clCreateBuffer(context,CL_MEM_WRITE_ONLY,3*sizeof(float)*solver.size,NULL,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    //solver.gpu_image = clCreateFromGLTexture2D(context,CL_MEM_WRITE_ONLY,GL_TEXTURE_2D,0,solver.tex_rgb,&err);
    //OPENCL_ERROR(err,"clCreateFromGLTexture")
    
    return 0; 
}    

int schrodinger2d_opencl_release_memory() {
    clReleaseMemObject(solver.gpu_U_re);
    clReleaseMemObject(solver.gpu_U_im);  
    clReleaseMemObject(solver.gpu_U1_re);
    clReleaseMemObject(solver.gpu_U1_im);
    clReleaseMemObject(solver.gpu_P);
    clReleaseMemObject(solver.gpu_Pmax);
    clReleaseMemObject(solver.gpu_A);
    clReleaseMemObject(solver.gpu_D);
    clReleaseMemObject(solver.gpu_E_re);
    clReleaseMemObject(solver.gpu_E_im);
    clReleaseMemObject(solver.gpu_F);
    clReleaseMemObject(solver.gpu_G);
    clReleaseMemObject(solver.gpu_beta_re);
    clReleaseMemObject(solver.gpu_beta_im);
    clReleaseMemObject(solver.gpu_beta_r2);
    clReleaseMemObject(solver.gpu_gamma_re);
    clReleaseMemObject(solver.gpu_gamma_im);
    clReleaseMemObject(solver.gpu_beta_p_re);
    clReleaseMemObject(solver.gpu_beta_p_im);
    clReleaseMemObject(solver.gpu_beta_p_r2);
    clReleaseMemObject(solver.gpu_gamma_p_re);
    clReleaseMemObject(solver.gpu_gamma_p_im);
    clReleaseMemObject(solver.gpu_map_r);
    clReleaseMemObject(solver.gpu_map_g);
    clReleaseMemObject(solver.gpu_map_b);
    clReleaseMemObject(solver.gpu_contours);
    clReleaseMemObject(solver.gpu_image);
    clReleaseMemObject(solver.gpu_sommets_z);
    clReleaseMemObject(solver.gpu_normales);
    return 0;   
}

int schrodinger2d_opencl_iterations(float ti, float tf) {
    int err;
    size_t global_work_size[1];
    size_t local_work_size[1];
    size_t global_work_offset[1] = {1};
    
    size_t proba_global_work_size[2];
    size_t proba_local_work_size[2];
    size_t proba_global_work_offset[2] = {1,1};
    
    size_t region[3];
    size_t origin[3] = {0,0,0};
    short ngx,ngy;
    float pmax;
    int k;
    
    
    err = clSetKernelArg(row_iteration,4,sizeof(cl_mem),(void*)&(solver.gpu_A));
    OPENCL_ERROR(err,"clSetKernelArg 4")
    err = clSetKernelArg(row_iteration,5,sizeof(cl_mem),(void*)&(solver.gpu_D));
    OPENCL_ERROR(err,"clSetKernelArg 5")
    err = clSetKernelArg(row_iteration,6,sizeof(cl_mem),(void*)&(solver.gpu_E_re));
    OPENCL_ERROR(err,"clSetKernelArg 6")
    err = clSetKernelArg(row_iteration,7,sizeof(cl_mem),(void*)&(solver.gpu_E_im));
    OPENCL_ERROR(err,"clSetKernelArg 7")
    err = clSetKernelArg(row_iteration,8,sizeof(cl_mem),(void*)&(solver.gpu_F));
    OPENCL_ERROR(err,"clSetKernelArg 8")
    err = clSetKernelArg(row_iteration,9,sizeof(cl_mem),(void*)&(solver.gpu_G));
    OPENCL_ERROR(err,"clSetKernelArg 9")
    err = clSetKernelArg(row_iteration,10,sizeof(cl_mem),(void*)&(solver.gpu_beta_re));
    OPENCL_ERROR(err,"clSetKernelArg 10")
    err = clSetKernelArg(row_iteration,11,sizeof(cl_mem),(void*)&(solver.gpu_beta_im));
    OPENCL_ERROR(err,"clSetKernelArg 11")
    err = clSetKernelArg(row_iteration,12,sizeof(cl_mem),(void*)&(solver.gpu_beta_r2));
    OPENCL_ERROR(err,"clSetKernelArg 12")
    err = clSetKernelArg(row_iteration,13,sizeof(cl_mem),(void*)&(solver.gpu_gamma_re));
    OPENCL_ERROR(err,"clSetKernelArg 13")
    err = clSetKernelArg(row_iteration,14,sizeof(cl_mem),(void*)&(solver.gpu_gamma_im));
    OPENCL_ERROR(err,"clSetKernelArg 14")
    err = clSetKernelArg(row_iteration,15,sizeof(solver.full_width),(void*)&(solver.full_width));
    OPENCL_ERROR(err,"clSetKernelArg 15")
    err = clSetKernelArg(row_iteration,16,sizeof(solver.full_height),(void*)&(solver.full_height));
    OPENCL_ERROR(err,"clSetKernelArg 16")
    err = clSetKernelArg(row_iteration,17,sizeof(solver.width),(void*)&(solver.width));
    OPENCL_ERROR(err,"clSetKernelArg 17")
    err = clSetKernelArg(row_iteration,18,sizeof(solver.height),(void*)&(solver.height));
    OPENCL_ERROR(err,"clSetKernelArg 18")
    
    err = clSetKernelArg(col_iteration,4,sizeof(cl_mem),(void*)&(solver.gpu_A));
    OPENCL_ERROR(err,"clSetKernelArg 4")
    err = clSetKernelArg(col_iteration,5,sizeof(cl_mem),(void*)&(solver.gpu_D));
    OPENCL_ERROR(err,"clSetKernelArg 5")
    err = clSetKernelArg(col_iteration,6,sizeof(cl_mem),(void*)&(solver.gpu_E_re));
    OPENCL_ERROR(err,"clSetKernelArg 6")
    err = clSetKernelArg(col_iteration,7,sizeof(cl_mem),(void*)&(solver.gpu_E_im));
    OPENCL_ERROR(err,"clSetKernelArg 7")
    err = clSetKernelArg(col_iteration,8,sizeof(cl_mem),(void*)&(solver.gpu_F));
    OPENCL_ERROR(err,"clSetKernelArg 8")
    err = clSetKernelArg(col_iteration,9,sizeof(cl_mem),(void*)&(solver.gpu_G));
    OPENCL_ERROR(err,"clSetKernelArg 9")
    err = clSetKernelArg(col_iteration,10,sizeof(cl_mem),(void*)&(solver.gpu_beta_p_re));
    OPENCL_ERROR(err,"clSetKernelArg 10")
    err = clSetKernelArg(col_iteration,11,sizeof(cl_mem),(void*)&(solver.gpu_beta_p_im));
    OPENCL_ERROR(err,"clSetKernelArg 11")
    err = clSetKernelArg(col_iteration,12,sizeof(cl_mem),(void*)&(solver.gpu_beta_p_r2));
    OPENCL_ERROR(err,"clSetKernelArg 12")
    err = clSetKernelArg(col_iteration,13,sizeof(cl_mem),(void*)&(solver.gpu_gamma_p_re));
    OPENCL_ERROR(err,"clSetKernelArg 13")
    err = clSetKernelArg(col_iteration,14,sizeof(cl_mem),(void*)&(solver.gpu_gamma_p_im));
    OPENCL_ERROR(err,"clSetKernelArg 14")
    err = clSetKernelArg(col_iteration,15,sizeof(solver.full_width),(void*)&(solver.full_width));
    OPENCL_ERROR(err,"clSetKernelArg 15")
    err = clSetKernelArg(col_iteration,16,sizeof(solver.full_height),(void*)&(solver.full_height));
    OPENCL_ERROR(err,"clSetKernelArg 16")
    err = clSetKernelArg(col_iteration,17,sizeof(solver.width),(void*)&(solver.width));
    OPENCL_ERROR(err,"clSetKernelArg 17")
    err = clSetKernelArg(col_iteration,18,sizeof(solver.height),(void*)&(solver.height));
    OPENCL_ERROR(err,"clSetKernelArg 18")
    
    
     
    
    local_work_size[0] = 1;
    solver.t = ti;
    while (solver.t<tf) {
        err = clSetKernelArg(row_iteration,0,sizeof(cl_mem),(void*)&(solver.gpu_U_re));
        OPENCL_ERROR(err,"clSetKernelArg 0")
        err = clSetKernelArg(row_iteration,1,sizeof(cl_mem),(void*)&(solver.gpu_U_im));
        OPENCL_ERROR(err,"clSetKernelArg 1")
        err = clSetKernelArg(row_iteration,2,sizeof(cl_mem),(void*)&(solver.gpu_U1_re));
        OPENCL_ERROR(err,"clSetKernelArg 2")
        err = clSetKernelArg(row_iteration,3,sizeof(cl_mem),(void*)&(solver.gpu_U1_im));
        OPENCL_ERROR(err,"clSetKernelArg 3")
        global_work_size[0] = solver.height;
        local_work_size[0] = 16;
        err= clEnqueueNDRangeKernel(commandQueue,row_iteration,1,global_work_offset,global_work_size,local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_1")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier");
         err = clSetKernelArg(col_iteration,0,sizeof(cl_mem),(void*)&(solver.gpu_U1_re));
        OPENCL_ERROR(err,"clSetKernelArg 0")
        err = clSetKernelArg(col_iteration,1,sizeof(cl_mem),(void*)&(solver.gpu_U1_im));
        OPENCL_ERROR(err,"clSetKernelArg 1")
        err = clSetKernelArg(col_iteration,2,sizeof(cl_mem),(void*)&(solver.gpu_U_re));
        OPENCL_ERROR(err,"clSetKernelArg 2")
        err = clSetKernelArg(col_iteration,3,sizeof(cl_mem),(void*)&(solver.gpu_U_im));
        OPENCL_ERROR(err,"clSetKernelArg 3")
        global_work_size[0] = solver.width;
        local_work_size[0] = 16;
        err= clEnqueueNDRangeKernel(commandQueue,col_iteration,1,global_work_offset,global_work_size,local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_2")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier");
        solver.t += solver.dt; 
    }
    err = clEnqueueReadBuffer(commandQueue, solver.gpu_U_re, CL_TRUE, 0,solver.full_size*sizeof(float), solver.U_re, 0, NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueReadBuffer U_re")
    err = clEnqueueReadBuffer(commandQueue, solver.gpu_U_im, CL_TRUE, 0,solver.full_size*sizeof(float), solver.U_im, 0, NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueReadBuffer U_im")
    
    
    err = clSetKernelArg(proba,0,sizeof(cl_mem),(void*)&(solver.gpu_U_re));
    OPENCL_ERROR(err,"clSetKernelArg 0")
    err = clSetKernelArg(proba,1,sizeof(cl_mem),(void*)&(solver.gpu_U_im));
    OPENCL_ERROR(err,"clSetKernelArg 1")
    err = clSetKernelArg(proba,2,sizeof(cl_mem),(void*)&(solver.gpu_P));
    OPENCL_ERROR(err,"clSetKernelArg 2")
    err = clSetKernelArg(proba,3,sizeof(cl_mem),(void*)&(solver.gpu_Pmax));
    OPENCL_ERROR(err,"clSetKernelArg 3")
    err = clSetKernelArg(proba,4,sizeof(solver.full_width),(void*)&(solver.full_width));
    OPENCL_ERROR(err,"clSetKernelArg 4")
    err = clSetKernelArg(proba,5,sizeof(solver.full_height),(void*)&(solver.full_height));
    OPENCL_ERROR(err,"clSetKernelArg 5")
    ngx = solver.width / PROBA_LOCAL_WORK_SIZE;
    ngy = solver.height / PROBA_LOCAL_WORK_SIZE;
    err = clSetKernelArg(proba,6,sizeof(ngx),(void*)&(ngx));
    OPENCL_ERROR(err,"clSetKernelArg 6")
    err = clSetKernelArg(proba,7,sizeof(ngy),(void*)&(ngy));
    OPENCL_ERROR(err,"clSetKernelArg 7")
    err = clSetKernelArg(proba,8,PROBA_LOCAL_WORK_SIZE*sizeof(float),NULL);
    OPENCL_ERROR(err,"clSetKernelArg 8")
    proba_local_work_size[0] = PROBA_LOCAL_WORK_SIZE;
    proba_global_work_size[0] = solver.width;
    proba_local_work_size[1] = PROBA_LOCAL_WORK_SIZE;
    proba_global_work_size[1] = solver.height;
    err= clEnqueueNDRangeKernel(commandQueue,proba,2,proba_global_work_offset,proba_global_work_size,proba_local_work_size,0,NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueNDRangeKernel_3")
    err = clEnqueueBarrier(commandQueue);
    OPENCL_ERROR(err,"clEnqueueBarrier");
    if (solver.normalize) {
        err = clEnqueueReadBuffer(commandQueue, solver.gpu_Pmax, CL_TRUE, 0,solver.Pmax_size*sizeof(float), solver.Pmax, 0, NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueReadBuffer")
        pmax = 0.0;
        for (k=0; k<solver.Pmax_size; k++) {
            if (solver.Pmax[k] > pmax) pmax = solver.Pmax[k];   
        }
        err = clSetKernelArg(norm_proba,0,sizeof(cl_mem),(void*)&(solver.gpu_P));
        OPENCL_ERROR(err,"clSetKernelArg 0")
        err = clSetKernelArg(norm_proba,1,sizeof(pmax),(void*)&(pmax));
        OPENCL_ERROR(err,"clSetKernelArg 1")
        err = clSetKernelArg(norm_proba,2,sizeof(solver.full_width),(void*)&(solver.full_width));
        OPENCL_ERROR(err,"clSetKernelArg 2")
        err = clSetKernelArg(norm_proba,3,sizeof(solver.full_height),(void*)&(solver.full_height));
        OPENCL_ERROR(err,"clSetKernelArg 3")
        err= clEnqueueNDRangeKernel(commandQueue,norm_proba,2,proba_global_work_offset,proba_global_work_size,proba_local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_4")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier");
    }
    if (solver.colormap) {
        proba_global_work_offset[0] = 0;
        proba_global_work_offset[1] = 0;
        
        
        err = clSetKernelArg(map_colors_img,0,sizeof(cl_mem),(void*)&(solver.gpu_P));
        OPENCL_ERROR(err,"clSetKernelArg 0")
        err = clSetKernelArg(map_colors_img,1,sizeof(cl_mem),(void*)&(solver.gpu_image));
        OPENCL_ERROR(err,"clSetKernelArg 1")
        err = clSetKernelArg(map_colors_img,2,sizeof(cl_mem),(void*)&(solver.gpu_map_r));
        OPENCL_ERROR(err,"clSetKernelArg 2")
        err = clSetKernelArg(map_colors_img,3,sizeof(cl_mem),(void*)&(solver.gpu_map_g));
        OPENCL_ERROR(err,"clSetKernelArg 3")
        err = clSetKernelArg(map_colors_img,4,sizeof(cl_mem),(void*)&(solver.gpu_map_b));
        OPENCL_ERROR(err,"clSetKernelArg 4")
        err = clSetKernelArg(map_colors_img,5,sizeof(cl_mem),(void*)&(solver.gpu_contours));
        OPENCL_ERROR(err,"clSetKernelArg 5")
        err = clSetKernelArg(map_colors_img,6,sizeof(solver.full_width),(void*)&(solver.full_width));
        OPENCL_ERROR(err,"clSetKernelArg 6")
        err = clSetKernelArg(map_colors_img,7,sizeof(solver.full_height),(void*)&(solver.full_height));
        OPENCL_ERROR(err,"clSetKernelArg 7")
        err = clSetKernelArg(map_colors_img,8,sizeof(solver.width),(void*)&(solver.width));
        OPENCL_ERROR(err,"clSetKernelArg 8")
        err = clSetKernelArg(map_colors_img,9,sizeof(solver.height),(void*)&(solver.height));
        OPENCL_ERROR(err,"clSetKernelArg 9")
        err = clSetKernelArg(map_colors_img,10,sizeof(solver.cont_r),(void*)&(solver.cont_r));
        OPENCL_ERROR(err,"clSetKernelArg 10")
        err = clSetKernelArg(map_colors_img,11,sizeof(solver.cont_r),(void*)&(solver.cont_g));
        OPENCL_ERROR(err,"clSetKernelArg 11")
        err = clSetKernelArg(map_colors_img,12,sizeof(solver.cont_r),(void*)&(solver.cont_b));
        OPENCL_ERROR(err,"clSetKernelArg 12")
        err = clSetKernelArg(map_colors_img,13,sizeof(solver.cont),(void*)&(solver.cont));
        OPENCL_ERROR(err,"clSetKernelArg 13")
        err= clEnqueueNDRangeKernel(commandQueue,map_colors_img,2,proba_global_work_offset,proba_global_work_size,proba_local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_5")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier");
        region[0] = solver.width;
        region[1] = solver.height;
        region[2] = 1;
        
        err = clEnqueueReadImage(commandQueue,solver.gpu_image,CL_TRUE,origin,region,0,0,solver.image,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueReadImage")
    }
    if (solver.surface) {
        err = clSetKernelArg(gen_surface,0,sizeof(cl_mem),(void*)&(solver.gpu_P));
        OPENCL_ERROR(err,"clSetKernelArg 0")
        err = clSetKernelArg(gen_surface,1,sizeof(cl_mem),(void*)&(solver.gpu_sommets_z));
        OPENCL_ERROR(err,"clSetKernelArg 1")
        err = clSetKernelArg(gen_surface,2,sizeof(cl_mem),(void*)&(solver.gpu_normales));
        OPENCL_ERROR(err,"clSetKernelArg 2")
        err = clSetKernelArg(gen_surface,3,sizeof(float),(void*)&(solver.surface_height));
        OPENCL_ERROR(err,"clSetKernelArg 3")
        err = clSetKernelArg(gen_surface,4,sizeof(float),(void*)&(solver.dx));
        OPENCL_ERROR(err,"clSetKernelArg 4")
        err = clSetKernelArg(gen_surface,5,sizeof(float),(void*)&(solver.dy));
        OPENCL_ERROR(err,"clSetKernelArg 5")
        err = clSetKernelArg(gen_surface,6,sizeof(float),(void*)&(solver.dxdy));
        OPENCL_ERROR(err,"clSetKernelArg 6")
        err = clSetKernelArg(gen_surface,7,sizeof(solver.full_width),(void*)&(solver.full_width));
        OPENCL_ERROR(err,"clSetKernelArg 7")
        err = clSetKernelArg(gen_surface,8,sizeof(solver.full_height),(void*)&(solver.full_height));
        OPENCL_ERROR(err,"clSetKernelArg 8")
        err = clSetKernelArg(gen_surface,9,sizeof(solver.width),(void*)&(solver.width));
        OPENCL_ERROR(err,"clSetKernelArg 9")
        err = clSetKernelArg(gen_surface,10,sizeof(solver.height),(void*)&(solver.height));
        OPENCL_ERROR(err,"clSetKernelArg 10")
        proba_global_work_offset[0] = 0;
        proba_global_work_offset[1] = 0;
        err= clEnqueueNDRangeKernel(commandQueue,gen_surface,2,proba_global_work_offset,proba_global_work_size,proba_local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_5")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier");
        err = clEnqueueReadBuffer(commandQueue, solver.gpu_sommets_z, CL_TRUE, 0,solver.size*sizeof(float), solver.sommets_z, 0, NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueReadBuffer sommets_z")
        err = clEnqueueReadBuffer(commandQueue, solver.gpu_normales, CL_TRUE, 0,3*solver.size*sizeof(float), solver.normales, 0, NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueReadBuffer normales")
        
    }
    err = clEnqueueReadBuffer(commandQueue, solver.gpu_P, CL_TRUE, 0,solver.full_size*sizeof(float), solver.P, 0, NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueReadBuffer P")
    return 0;
}


#endif

