#include <stdio.h>
#include <stdlib.h>
#include "shaders.h"
#include <GL/glew.h>

const char* vertexShader_basic[] = {
       "attribute vec2 aVertexPosition;",
       "void main() {",
            "gl_Position = vec4(aVertexPosition,0.0,1.0);",
        "}"
};
const int vertexShader_basic_length = 4;

const char* fragmentShader_basic[] = {
    "uniform vec4 uColor;",
    "void main() {",
        "gl_FragColor = uColor;",
    "}"
};
const int fragmentShader_basic_length = 4;

const char* vertexShader_pixmap[] = {
    "attribute vec2 aVertexPosition;",
     "attribute vec2 aTexCoord;",
     "varying vec2 tex_coord;",
      "void main() {",
                "gl_Position = vec4(aVertexPosition, 0.0,1.0);",
                "tex_coord = aTexCoord;",
       "}" 
};
const int vertexShader_pixmap_length = 7;

const char* fragmentShader_pixmap[] = {
    "uniform sampler2D uSampler;",
    "in vec2 tex_coord;",
     "void main() {",
            "float tex = texture(uSampler,tex_coord);",
            "gl_FragColor = vec4(tex,tex,tex,1.0);",
     "}"  
};
const int fragmentShader_pixmap_length = 6;

const char* fragmentShader_image[] = {
    "uniform sampler2D uSampler_r;",
    "uniform sampler2D uSampler_g;",
    "uniform sampler2D uSampler_b;",
    "in vec2 tex_coord;",
    "void main() {",
        "float r = texture(uSampler_r,tex_coord);",
        "float g = texture(uSampler_g,tex_coord);",
        "float b = texture(uSampler_b,tex_coord);",
        "gl_FragColor = vec4(r,g,b,1.0);",
    "}"
};
const int fragmentShader_image_length = 10;

const char* fragmentShader_image_rgb[] = {
    "#version 130\n",
    "uniform sampler2D uSampler;",
    "in vec2 tex_coord;",
    "void main() {",
        "vec4 pixel = texture(uSampler,tex_coord);",
        "gl_FragColor = vec4(pixel);",
    "}",
};
const int fragmentShader_image_rgb_length = 7;

const char* vertexShader_surface[] = {
    "attribute vec2 aVertexXY;",
    "attribute float aVertexZ;",
    "attribute vec3 aVertexNormal;",
    "attribute vec4 aColor;",
    "uniform mat4 uPMVMatrix;",
    "uniform mat4 uMVMatrix;",
    "uniform mat3 uNormalMatrix;",
    "uniform vec4 uLightPosition;",
    "uniform vec3 uKdiffuse;",
    "uniform vec3 uLightIntensity;",
    "uniform vec3 uKambient;",
    "varying vec3 color;",
    "void main() {",
        "vec4 vertex = vec4(aVertexXY,aVertexZ,1.0);",
        "vec3 tnorm = uNormalMatrix*normalize(aVertexNormal);",
        "vec4 eyeCoords = uMVMatrix*vertex;",
        "vec3 s;",
        "if (uLightPosition.w==0.0) s = normalize(vec3(uLightPosition));",
        "else s = normalize(vec3(uLightPosition-eyeCoords));",
        "gl_Position = uPMVMatrix*vertex;",
        "color = aColor*uLightIntensity*(uKdiffuse*max(dot(s,tnorm),0.0)+uKambient);",
    "}"
};
const int vertexShader_surface_length = 22;

const char* fragmentShader_surface[] = {
    "in vec3 color;",
    "void main() {",
        "gl_FragColor = vec4(color,1.0);",
    "}"
};
const int fragmentShader_surface_length = 4;

const char* vertexShader_pixmap_modelview[] = {
     "attribute vec2 aVertexPosition;",
     "attribute vec2 aTexCoord;",
     "uniform mat4 uPMVMatrix;",
     "varying vec2 tex_coord;",
     "void main() {",
                "gl_Position = uPMVMatrix*vec4(aVertexPosition, 0.01,1);",
                "tex_coord = aTexCoord;",
      "}" 
};
const int vertexShader_pixmap_modelview_length = 8;

GLuint get_gl_program(const char** vertexShader, const int vertexShader_length, const char** fragmentShader, const int fragmentShader_length) {
    GLuint shader;
    GLint params;
    char *log;
    GLsizei length;
    GLuint program;
    program = glCreateProgram();
    shader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(shader,vertexShader_length,vertexShader,NULL);
    glCompileShader(shader);
    glGetShaderiv(shader,GL_COMPILE_STATUS,&params);
    if (params!=GL_TRUE) {
        printf("Erreur de compilation vertex shader\n");
        glGetShaderiv(shader,GL_INFO_LOG_LENGTH,&params);
        log = (char*) malloc(params+1);
        glGetShaderInfoLog(shader,params,&length,log);
        printf("%s\n",log);
        free(log);
    }
    glAttachShader(program,shader);
    shader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(shader,fragmentShader_length,fragmentShader,NULL);
    glCompileShader(shader);
    glGetShaderiv(shader,GL_COMPILE_STATUS,&params);
    if (params!=GL_TRUE) {
        printf("Erreur de compilation fragment shader\n");
        glGetShaderiv(shader,GL_INFO_LOG_LENGTH,&params);
        log = (char*) malloc(params+1);
        glGetShaderInfoLog(shader,params,&length,log);
        printf("%s\n",log);
        free(log);
    }
    glAttachShader(program,shader);
    glLinkProgram(program);
    return program;
}


GLuint get_gl_program_basic() {
    return get_gl_program(vertexShader_basic,vertexShader_basic_length,fragmentShader_basic,fragmentShader_basic_length);
    
}

GLuint get_gl_program_pixmap() {
    return get_gl_program(vertexShader_pixmap,vertexShader_pixmap_length,fragmentShader_pixmap,fragmentShader_pixmap_length);
}

GLuint get_gl_program_image() {
    return get_gl_program(vertexShader_pixmap,vertexShader_pixmap_length,fragmentShader_image,fragmentShader_image_length);
}

GLuint get_gl_program_image_rgb() {
    return get_gl_program(vertexShader_pixmap,vertexShader_pixmap_length,fragmentShader_image_rgb,fragmentShader_image_rgb_length);
}

GLuint get_gl_program_surface() {
    return get_gl_program(vertexShader_surface,vertexShader_surface_length,fragmentShader_surface,fragmentShader_surface_length);   
}

GLuint get_gl_program_pixmap_modelview() {
     return get_gl_program(vertexShader_pixmap_modelview,vertexShader_pixmap_modelview_length,fragmentShader_pixmap,fragmentShader_pixmap_length);  
}



void draw_basic_triangle(GLuint program) {
     float color[4] = {1.0,0.1,0.7,1.0};
     const GLfloat positions[] = {
          0.0f, 0.0f,
          0.0f, 1.0f,
          1.0f, 0.0f
     };
     GLint aVertexPosition_location, uColor_location;
     GLuint buffer;
     
     glUseProgram(program);
     GL_ERROR
     aVertexPosition_location = glGetAttribLocation(program,"aVertexPosition");
     GL_ERROR
     uColor_location = glGetUniformLocation(program,"uColor");
     GL_ERROR
     glUniform4fv(uColor_location,1,color);
     
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(positions),positions,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aVertexPosition_location);
     glVertexAttribPointer(aVertexPosition_location,2,GL_FLOAT,GL_FALSE,0,0);
     GL_ERROR
     glDrawArrays(GL_TRIANGLES,0,3);
     GL_ERROR
     glBindBuffer(GL_ARRAY_BUFFER,0);
     
}

void draw_pixmap(GLuint program, int width, int height, float *data ) {
    const GLfloat positions[] = {
          -1.0f, -1.0f,
          1.0f, -1.0f,
          -1.0f, 1.0f,
          -1.0f, 1.0f,
          1.0f,-1.0f,
          1.0f,1.0f
     };
     const GLfloat texcoords[] = {
          0.0f,0.0f,
          1.0f,0.0f,
          0.0f,1.0f,
          0.0f,1.0f,
          1.0f,0.0f,
          1.0f,1.0f
     };
     GLint aVertexPosition_location, aTexCoord_location, usampler_location;
     GLuint buffer;
     GLuint texture;
     glUseProgram(program);
     aVertexPosition_location = glGetAttribLocation(program,"aVertexPosition");
     aTexCoord_location = glGetAttribLocation(program,"aTexCoord");
     usampler_location = glGetUniformLocation(program,"uSampler");
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(positions),positions,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aVertexPosition_location);
     glVertexAttribPointer(aVertexPosition_location,2,GL_FLOAT,GL_FALSE,0,0);
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(texcoords),texcoords,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aTexCoord_location);
     glVertexAttribPointer(aTexCoord_location,2,GL_FLOAT,GL_FALSE,0,0);
     glGenTextures(1,&texture);
     glBindTexture(GL_TEXTURE_2D,texture);
     glPixelStoref(GL_UNPACK_ALIGNMENT,1);
     glTexImage2D(GL_TEXTURE_2D,0,GL_RED,width,height,0,GL_RED,GL_FLOAT,data);
     glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
     glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
     glActiveTexture(GL_TEXTURE0);
     glUniform1i(usampler_location,0);
     glDrawArrays(GL_TRIANGLES,0,6);
     glBindTexture(GL_TEXTURE_2D,0);
     glBindBuffer(GL_ARRAY_BUFFER,0);
}

void draw_image(GLuint program, GLuint tex_r, GLuint tex_g, GLuint tex_b) {
    const GLfloat positions[] = {
          -1.0f, -1.0f,
          1.0f, -1.0f,
          -1.0f, 1.0f,
          -1.0f, 1.0f,
          1.0f,-1.0f,
          1.0f,1.0f
     };
     const GLfloat texcoords[] = {
          0.0f,0.0f,
          1.0f,0.0f,
          0.0f,1.0f,
          0.0f,1.0f,
          1.0f,0.0f,
          1.0f,1.0f
     };
     GLint aVertexPosition_location, aTexCoord_location, usampler_r_location, usampler_g_location, usampler_b_location;
     GLuint buffer;
     glUseProgram(program);
     aVertexPosition_location = glGetAttribLocation(program,"aVertexPosition");
     aTexCoord_location = glGetAttribLocation(program,"aTexCoord");
     usampler_r_location = glGetUniformLocation(program,"uSampler_r");
     usampler_g_location = glGetUniformLocation(program,"uSampler_g");
     usampler_b_location = glGetUniformLocation(program,"uSampler_b");
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(positions),positions,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aVertexPosition_location);
     glVertexAttribPointer(aVertexPosition_location,2,GL_FLOAT,GL_FALSE,0,0);
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(texcoords),texcoords,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aTexCoord_location);
     glVertexAttribPointer(aTexCoord_location,2,GL_FLOAT,GL_FALSE,0,0);
     glActiveTexture(GL_TEXTURE0);
     glBindTexture(GL_TEXTURE_2D,tex_r);
     glUniform1i(usampler_r_location,0);
     glActiveTexture(GL_TEXTURE1);
     glBindTexture(GL_TEXTURE_2D,tex_g);
     glUniform1i(usampler_g_location,1);
     glActiveTexture(GL_TEXTURE2);
     glBindTexture(GL_TEXTURE_2D,tex_b);
     glUniform1i(usampler_b_location,2);
     glDrawArrays(GL_TRIANGLES,0,6);
     glBindTexture(GL_TEXTURE_2D,0);
     glBindBuffer(GL_ARRAY_BUFFER,0);
}

void draw_image_rgb(GLuint program, GLuint texture) {
    const GLfloat positions[] = {
          -1.0f, -1.0f,
          1.0f, -1.0f,
          -1.0f, 1.0f,
          -1.0f, 1.0f,
          1.0f,-1.0f,
          1.0f,1.0f
     };
     const GLfloat texcoords[] = {
          0.0f,0.0f,
          1.0f,0.0f,
          0.0f,1.0f,
          0.0f,1.0f,
          1.0f,0.0f,
          1.0f,1.0f
     };
     GLint aVertexPosition_location, aTexCoord_location, usampler_location;
     GLuint buffer;
     glUseProgram(program);
     aVertexPosition_location = glGetAttribLocation(program,"aVertexPosition");
     aTexCoord_location = glGetAttribLocation(program,"aTexCoord");
     usampler_location = glGetUniformLocation(program,"uSampler");
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(positions),positions,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aVertexPosition_location);
     glVertexAttribPointer(aVertexPosition_location,2,GL_FLOAT,GL_FALSE,0,0);
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(texcoords),texcoords,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aTexCoord_location);
     glVertexAttribPointer(aTexCoord_location,2,GL_FLOAT,GL_FALSE,0,0);
     glActiveTexture(GL_TEXTURE0);
     glBindTexture(GL_TEXTURE_2D,texture);
     glUniform1i(usampler_location,0);
     glDrawArrays(GL_TRIANGLES,0,6);
     glBindTexture(GL_TEXTURE_2D,0);
     glBindBuffer(GL_ARRAY_BUFFER,0);
}

void draw_pixmap_modelview(GLuint program, GLuint texture, float xmin, float xmax, float ymin, float ymax, float *PMV) {
    GLfloat positions[] = {
          -1.0f, -1.0f,
          1.0f, -1.0f,
          -1.0f, 1.0f,
          -1.0f, 1.0f,
          1.0f,-1.0f,
          1.0f,1.0f
     };
     const GLfloat texcoords[] = {
          0.0f,0.0f,
          1.0f,0.0f,
          0.0f,1.0f,
          0.0f,1.0f,
          1.0f,0.0f,
          1.0f,1.0f
     };
     GLint aVertexPosition_location, aTexCoord_location, usampler_location,uPMVMatrix_location;
     GLuint buffer;
     positions[0] = xmin; positions[1] = ymin;
     positions[2] = xmax; positions[3] = ymin;
     positions[4] = xmin; positions[5] = ymax;
     positions[6] = xmin; positions[7] = ymax;
     positions[8] = xmax; positions[9] = ymin;
     positions[10] = xmax; positions[11] = ymax;
     glUseProgram(program);
     aVertexPosition_location = glGetAttribLocation(program,"aVertexPosition");
     aTexCoord_location = glGetAttribLocation(program,"aTexCoord");
     usampler_location = glGetUniformLocation(program,"uSampler");
     uPMVMatrix_location = glGetUniformLocation(program,"uPMVMatrix");
     glUniformMatrix4fv(uPMVMatrix_location,1,GL_FALSE,PMV);
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(positions),positions,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aVertexPosition_location);
     glVertexAttribPointer(aVertexPosition_location,2,GL_FLOAT,GL_FALSE,0,0);
     glGenBuffers(1,&buffer);
     glBindBuffer(GL_ARRAY_BUFFER,buffer);
     glBufferData(GL_ARRAY_BUFFER,sizeof(texcoords),texcoords,GL_STATIC_DRAW);
     glEnableVertexAttribArray(aTexCoord_location);
     glVertexAttribPointer(aTexCoord_location,2,GL_FLOAT,GL_FALSE,0,0);
     glActiveTexture(GL_TEXTURE0);
     glBindTexture(GL_TEXTURE_2D,texture);
     glUniform1i(usampler_location,0);
     glDrawArrays(GL_TRIANGLES,0,6);
     glBindTexture(GL_TEXTURE_2D,0);
     glBindBuffer(GL_ARRAY_BUFFER,0);
}

void draw_surface(GLuint program, int nb_sommets, GLuint buffer_indices, GLuint buffer_sommets_xy, GLuint buffer_sommets_z, GLuint buffer_normales, GLuint buffer_colors, float *PMV, float *MV, float *N, float *LightPosition, float *Kdiffuse, float *LightIntensity, float *Kambient) {
     GLuint aVertexXY,aVertexZ,aVertexNormal,aColor;
     GLuint uPMVMatrix, uMVMatrix, uNormalMatrix,uLightPosition,uKdiffuse,uLightIntensity,uKambient;
     glUseProgram(program);
     aVertexXY = glGetAttribLocation(program,"aVertexXY");
     aVertexZ = glGetAttribLocation(program,"aVertexZ");
     aVertexNormal = glGetAttribLocation(program,"aVertexNormal");
     aColor = glGetAttribLocation(program,"aColor");
     uPMVMatrix = glGetUniformLocation(program,"uPMVMatrix");
     uMVMatrix = glGetUniformLocation(program,"uMVMatrix");
     uNormalMatrix = glGetUniformLocation(program,"uNormalMatrix");
     uLightPosition = glGetUniformLocation(program,"uLightPosition");
     uKdiffuse = glGetUniformLocation(program,"uKdiffuse");
     uLightIntensity = glGetUniformLocation(program,"uLightIntensity");
     uKambient = glGetUniformLocation(program,"uKambient");
     
     glBindBuffer(GL_ARRAY_BUFFER,buffer_sommets_xy);
     glEnableVertexAttribArray(aVertexXY);
     glVertexAttribPointer(aVertexXY,2,GL_FLOAT,GL_FALSE,0,0);
     glBindBuffer(GL_ARRAY_BUFFER,buffer_sommets_z);
     glEnableVertexAttribArray(aVertexZ);
     glVertexAttribPointer(aVertexZ,1,GL_FLOAT,GL_FALSE,0,0);
     glBindBuffer(GL_ARRAY_BUFFER,buffer_normales);
     glEnableVertexAttribArray(aVertexNormal);
     glVertexAttribPointer(aVertexNormal,3,GL_FLOAT,GL_FALSE,0,0);
     glBindBuffer(GL_ARRAY_BUFFER,buffer_colors);
     glEnableVertexAttribArray(aColor);
     glVertexAttribPointer(aColor,4,GL_FLOAT,GL_FALSE,0,0);
     glUniformMatrix4fv(uPMVMatrix,1,GL_FALSE,PMV);
     glUniformMatrix4fv(uMVMatrix,1,GL_FALSE,MV);
     glUniformMatrix3fv(uNormalMatrix,1,GL_FALSE,N);
     glUniform4fv(uLightPosition,1,LightPosition);
     glUniform3fv(uKdiffuse,1,Kdiffuse);
     glUniform3fv(uLightIntensity,1,LightIntensity);
     glUniform3fv(uKambient,1,Kambient);
     glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,buffer_indices);
     glDrawElements(GL_TRIANGLES,nb_sommets,GL_UNSIGNED_INT,0);
     glBindBuffer(GL_ARRAY_BUFFER,0);
     glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
}

