#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "transfo.h"

#define DEG2RAD 1.74532925e-2


void matrix_prod(float *a, float *b, float *c) {
    c[0] = a[0]*b[0]+a[4]*b[1]+a[8]*b[2]+a[12]*b[3];
    c[1] = a[1]*b[0]+a[5]*b[1]+a[9]*b[2]+a[13]*b[3];
    c[2] = a[2]*b[0]+a[6]*b[1]+a[10]*b[2]+a[14]*b[3];
    c[3] = a[3]*b[0]+a[7]*b[1]+a[11]*b[2]+a[15]*b[3];
    
    c[4] = a[0]*b[4]+a[4]*b[5]+a[8]*b[6]+a[12]*b[7];
    c[5] = a[1]*b[4]+a[5]*b[5]+a[9]*b[6]+a[13]*b[7];
    c[6] = a[2]*b[4]+a[6]*b[5]+a[10]*b[6]+a[14]*b[7];
    c[7] = a[3]*b[4]+a[7]*b[5]+a[11]*b[6]+a[15]*b[7];
    
    c[8]  = a[0]*b[8]+a[4]*b[9]+a[8]*b[10]+a[12]*b[11];
    c[9]  = a[1]*b[8]+a[5]*b[9]+a[9]*b[10]+a[13]*b[11];
    c[10] = a[2]*b[8]+a[6]*b[9]+a[10]*b[10]+a[14]*b[11];
    c[11] = a[3]*b[8]+a[7]*b[9]+a[11]*b[10]+a[15]*b[11];
    
    c[12] = a[0]*b[12]+a[4]*b[13]+a[8]*b[14]+a[12]*b[15];
    c[13] = a[1]*b[12]+a[5]*b[13]+a[9]*b[14]+a[13]*b[15];
    c[14] = a[2]*b[12]+a[6]*b[13]+a[10]*b[14]+a[14]*b[15];
    c[15] = a[3]*b[12]+a[7]*b[13]+a[11]*b[14]+a[15]*b[15];
}

void matrix_copy(float *a, float *b) {
    int i;
    for (i=0; i<16; i++) b[i] = a[i];
}

void matrix_id(float *a) {
    a[0] = a[5] = a[10] = a[15] = 1.0;
    a[1] = a[2] = a[3] = 0.0;
    a[4] = a[6] = a[7] = 0.0;
    a[8] = a[9] = a[11] = 0.0;
    a[12] = a[13] = a[14] = 0.0;
}

void transfo_new(Transformation **trans) {
    *trans = (Transformation*)malloc(sizeof(Transformation));
    (*trans)->last = NULL;
    (*trans)->M = (float*)malloc(16*sizeof(float));
    (*trans)->invM = (float*)malloc(16*sizeof(float));
    matrix_id((*trans)->M);
    matrix_id((*trans)->invM);
}

void transfo_delete(Transformation *trans) {
    free(trans->M);
    free(trans->invM);
    if (trans->last!=NULL) transfo_delete(trans->last);
    free(trans);
}

void transfo_init(Transformation *trans) {
    matrix_id(trans->M);
    matrix_id(trans->invM);
}

void transfo_copy(Transformation *trans, Transformation *modele) {
    matrix_copy(modele->M,trans->M);
    matrix_copy(modele->invM,trans->invM);
}

void transfo_compose(Transformation *trans, Transformation *t) {
    float *c;
    c = (float*)malloc(16*sizeof(float));
    matrix_prod(trans->M,t->M,c);
    matrix_copy(c,trans->M);
    matrix_prod(t->invM,trans->invM,c);
    matrix_copy(c,trans->invM);
    free(c);
}

void transfo_save(Transformation *trans) {
    Transformation *t;
    transfo_new(&t);
    matrix_copy(trans->M,t->M);
    matrix_copy(trans->invM,t->invM);
    t->last = trans->last;
    trans->last = t;
}

void transfo_restore(Transformation *trans) {
    Transformation *t;
    if (trans->last==NULL) return;
    t = (Transformation*) (trans->last);
    trans->last = t->last;
    matrix_copy(t->M,trans->M);
    matrix_copy(t->invM,trans->invM);
}

void transfo_translate(Transformation *trans, float tx, float ty, float tz) {
    Transformation *T;
    transfo_new(&T);
    T->M[12] = tx; T->M[13] = ty; T->M[14] = tz;
    T->invM[12] = -tx; T->invM[13] = -ty; T->invM[14] = -tz;
    transfo_compose(trans,T);
    transfo_delete(T);
}

void transfo_rotate_z_rad(Transformation *trans, float a) {
    Transformation *T;
    float c,s;
    transfo_new(&T);
    c = cos(a); s = sin(a);
    T->M[0] = c; T->M[1] = s; T->M[4] = -s; T->M[5] = c;
    T->invM[0] = c; T->invM[1] = -s; T->invM[4] = s; T->invM[5] = c;
    transfo_compose(trans,T);
    transfo_delete(T);  
}

void transfo_rotate_z(Transformation *trans, float a) {
     transfo_rotate_z_rad(trans,a*DEG2RAD); 
}

void transfo_rotate_x_rad(Transformation *trans, float a) {
    Transformation *T;
    float c,s;
    transfo_new(&T);
    c = cos(a); s = sin(a);
    T->M[5] = c; T->M[6] = s; T->M[9] = -s; T->M[10] = c;
    T->invM[5] = c; T->invM[6] = -s; T->invM[9] = s; T->invM[10] = c;
    transfo_compose(trans,T);
    transfo_delete(T);  
}

void transfo_rotate_x(Transformation *trans, float a) {
     transfo_rotate_x_rad(trans,a*DEG2RAD); 
}

void transfo_rotate_y_rad(Transformation *trans, float a) {
    Transformation *T;
    float c,s;
    transfo_new(&T);
    c = cos(a); s = sin(a);
    T->M[0] = c; T->M[2] = -s; T->M[8] = s; T->M[10] = c;
    T->invM[0] = c; T->invM[2] = s; T->invM[8] = -s; T->invM[10] = c;
    transfo_compose(trans,T);
    transfo_delete(T);  
}

void transfo_rotate_y(Transformation *trans, float a) {
     transfo_rotate_y_rad(trans,a*DEG2RAD); 
}

void transfo_scale(Transformation *trans, float sx, float sy, float sz) {
    Transformation *T;
    transfo_new(&T);
    T->M[0] = sx; T->M[5] = sy; T->M[10] = sz;
    T->invM[0] = 1.0/sx; T->invM[5] = 1.0/sy; T->invM[10] = 1.0/sz;
    transfo_compose(trans,T);
    transfo_delete(T);  
}

void projection_frustum(Transformation *trans, float width, float height, float near, float far) {
    float a,b,depth,c,d;
    a = -2.0*near/width;
    b = -2.0*near/height;
    depth = near-far;
    c=(near+far)/depth;
    d=-2.0*near*far/depth;
    trans->M[0] = a; trans->M[5] = b;
    trans->M[10] = c;
    trans->M[11] = -1.0;
    trans->M[14] = d;
    trans->M[15] = 0.0;
}

void projection_ortho(Transformation *trans, float width, float height, float near, float far) {
    float a,b,depth,c,d;
    a = 2.0/width;
    b = 2.0/height;
    depth = near-far;
    c=2.0/depth;
    d=-(near+far)/depth;
    trans->M[0] = a; trans->M[5] = b;
    trans->M[10] = c;
    trans->M[14] = d;
    
}

void transfo_get_normal_matrix(Transformation *trans, float *N) {
    float *I = trans->invM;
    N[0] = I[0]; N[1] = I[4]; N[2] = I[8];
    N[3] = I[1]; N[4] = I[5]; N[5] = I[9];
    N[6] = I[2]; N[7] = I[6]; N[8] = I[10];
}










